#!/usr/bin/env python3


import argparse
import asyncio
import functools
import inspect
import shutil
import sys
from typing import get_type_hints

import websockets
from aiologger.levels import LogLevel
from aiorun import run
from dataclasses_jsonschema import ValidationError
from websockets.server import WebSocketServerProtocol as WsClient

import arcor2.helpers as hlp
import arcor2_arserver
import arcor2_arserver_data
import arcor2_execution_data
from arcor2 import env, json, ws_server
from arcor2.clients import aio_scene_service as scene_srv
from arcor2.data import compile_json_schemas, events, rpc
from arcor2.exceptions import Arcor2Exception
from arcor2.parameter_plugins.utils import known_parameter_types
from arcor2_arserver import events as server_events
from arcor2_arserver import execution as exe
from arcor2_arserver import globals as glob
from arcor2_arserver import logger, models
from arcor2_arserver import notifications as notif
from arcor2_arserver import objects_actions as osa
from arcor2_arserver import rpc as srpc_callbacks
from arcor2_arserver import scene, settings
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.lock.notifications import run_lock_notification_worker
from arcor2_arserver_data import events as evts
from arcor2_arserver_data import rpc as srpc
from arcor2_arserver_data.rpc import objects as obj_rpc
from arcor2_execution_data import EVENTS as EXE_EVENTS
from arcor2_execution_data import EXPOSED_RPCS
from arcor2_execution_data import RPCS as EXE_RPCS
from arcor2_runtime import events as runtime_events


async def handle_manager_incoming_messages(manager_client) -> None:

    event_mapping: dict[str, type[events.Event]] = {evt.__name__: evt for evt in EXE_EVENTS}
    rpc_mapping: dict[str, type[rpc.common.RPC]] = {r.__name__: r for r in EXE_RPCS}

    try:

        async for message in manager_client:

            msg = json.loads(message)

            if not isinstance(msg, dict):
                continue

            if "event" in msg:

                if glob.USERS.interfaces:
                    await asyncio.gather(
                        *[ws_server.send_json_to_client(intf, message) for intf in glob.USERS.interfaces]
                    )

                try:
                    evt = event_mapping[msg["event"]].from_dict(msg)
                except ValidationError as e:
                    logger.error("Invalid event: {}, error: {}".format(msg, e))
                    continue

                if isinstance(evt, events.PackageInfo):
                    glob.PACKAGE_INFO = evt.data
                elif isinstance(evt, events.PackageState):
                    glob.PACKAGE_STATE = evt.data

                    if evt.data.state == events.PackageState.Data.StateEnum.STOPPED:

                        if not glob.TEMPORARY_PACKAGE:
                            # after (ordinary) package is finished, show list of packages
                            glob.MAIN_SCREEN = evts.c.ShowMainScreen.Data(
                                evts.c.ShowMainScreen.Data.WhatEnum.PackagesList
                            )
                            await notif.broadcast_event(
                                evts.c.ShowMainScreen(
                                    evts.c.ShowMainScreen.Data(
                                        evts.c.ShowMainScreen.Data.WhatEnum.PackagesList, evt.data.package_id
                                    )
                                )
                            )

                        # temporary package is handled elsewhere
                        server_events.package_stopped.set()
                        server_events.package_started.clear()

                        glob.ACTION_STATE_BEFORE = None
                        glob.PACKAGE_INFO = None

                    else:
                        server_events.package_stopped.clear()
                        server_events.package_started.set()

                elif isinstance(evt, events.ActionStateBefore):
                    glob.ACTION_STATE_BEFORE = evt.data

            elif "response" in msg:

                # TODO handle potential errors
                rpc_cls = rpc_mapping[msg["response"]]
                resp = rpc_cls.Response.from_dict(msg)
                exe.MANAGER_RPC_RESPONSES[resp.id].put_nowait(resp)

    except websockets.exceptions.ConnectionClosed as e:
        logger.error(f"Connection to manager closed. {str(e)}")


async def _initialize_server() -> None:

    exe_version = await exe.manager_request(rpc.common.Version.Request(exe.get_id()))
    assert isinstance(exe_version, rpc.common.Version.Response)
    if not exe_version.result:
        raise Arcor2Exception("Failed to get Execution version.")

    assert exe_version.data is not None

    """
    Following check is especially useful when running server/execution in Docker containers.
    Then it might easily happen that one tries to use different versions together.
    """
    try:
        hlp.check_compatibility(arcor2_execution_data.version(), exe_version.data.version)
    except Arcor2Exception as e:
        raise Arcor2Exception("ARServer/Execution uses different versions of arcor2_execution_data.") from e

    while True:  # wait until Project service becomes available
        try:
            await storage.initialize_module()
            break
        except storage.ProjectServiceException as e:
            logger.error(f"Failed to communicate with Project service. {str(e)}")
            await asyncio.sleep(1)

    while True:
        try:
            if await scene_srv.started():
                logger.warn("Scene already started, attempting to stop it...")
            await scene_srv.stop()  # if not started, call it anyway, it should be instant
            break
        except scene_srv.SceneServiceException as e:
            logger.error(f"Failed to communicate with the Scene service. {str(e)}")
            await asyncio.sleep(1)

    await osa.get_object_types()

    bound_handler = functools.partial(
        ws_server.server,
        logger=logger,
        register=register,
        unregister=unregister,
        rpc_dict=RPC_DICT,
        event_dict=EVENT_DICT,
        verbose=glob.VERBOSE,
    )

    if __debug__:
        logger.warn("Development mode. The service will shutdown on any unhandled exception.")

    logger.info(f"ARServer {arcor2_arserver.version()} " f"(API version {arcor2_arserver_data.version()}) initialized.")
    await asyncio.wait([websockets.server.serve(bound_handler, "0.0.0.0", glob.PORT)])
    asyncio.create_task(run_lock_notification_worker())


async def list_meshes_cb(req: obj_rpc.ListMeshes.Request, ui: WsClient) -> obj_rpc.ListMeshes.Response:
    return obj_rpc.ListMeshes.Response(data=await storage.get_meshes())


async def register(websocket: WsClient) -> None:

    logger.info("Registering new ui")
    glob.USERS.add_interface(websocket)

    if glob.LOCK.project:
        assert glob.LOCK.scene
        await notif.event(
            websocket, evts.p.OpenProject(evts.p.OpenProject.Data(glob.LOCK.scene.scene, glob.LOCK.project.project))
        )
    elif glob.LOCK.scene:
        await notif.event(websocket, evts.s.OpenScene(evts.s.OpenScene.Data(glob.LOCK.scene.scene)))
    elif glob.PACKAGE_INFO:

        # this can't be done in parallel - ui expects this order of events
        await websocket.send(events.PackageState(glob.PACKAGE_STATE).to_json())
        await websocket.send(events.PackageInfo(glob.PACKAGE_INFO).to_json())

        if glob.ACTION_STATE_BEFORE:
            await websocket.send(events.ActionStateBefore(glob.ACTION_STATE_BEFORE).to_json())
    else:
        assert glob.MAIN_SCREEN
        await notif.event(websocket, evts.c.ShowMainScreen(glob.MAIN_SCREEN))

    if glob.LOCK.project or glob.LOCK.scene:
        await notif.event(websocket, scene.get_scene_state())


async def unregister(websocket: WsClient) -> None:

    try:
        user_name = glob.USERS.user_name(websocket)
    except Arcor2Exception:
        logger.info("Unregistering ui")
    else:
        await glob.LOCK.schedule_auto_release(user_name)
        logger.info(f"Unregistering ui {user_name}")

    glob.USERS.logout(websocket)

    logger.debug(f"Known user names: {glob.USERS.user_names}")

    for registered_uis in glob.ROBOT_JOINTS_REGISTERED_UIS.values():
        if websocket in registered_uis:
            registered_uis.remove(websocket)
    for registered_uis in glob.ROBOT_EEF_REGISTERED_UIS.values():
        if websocket in registered_uis:
            registered_uis.remove(websocket)


async def system_info_cb(req: srpc.c.SystemInfo.Request, ui: WsClient) -> srpc.c.SystemInfo.Response:

    resp = srpc.c.SystemInfo.Response()
    resp.data = resp.Data(
        arcor2_arserver.version(),
        arcor2_arserver_data.version(),
        known_parameter_types(),
        {key for key in RPC_DICT.keys()},
    )
    return resp


RPC_DICT: ws_server.RPC_DICT_TYPE = {srpc.c.SystemInfo.__name__: (srpc.c.SystemInfo, system_info_cb)}

# discovery of RPC callbacks
# TODO refactor it into arcor2 package (to be used by arcor2_execution)
for _, rpc_module in inspect.getmembers(srpc_callbacks, inspect.ismodule):
    for rpc_cb_name, rpc_cb in inspect.getmembers(rpc_module):

        if not rpc_cb_name.endswith("_cb"):
            continue

        hints = get_type_hints(rpc_cb)

        req_cls = hints["req"]
        rpc_cls = getattr(sys.modules[req_cls.__module__], req_cls.__qualname__.split(".")[0])

        RPC_DICT[rpc_cls.__name__] = (rpc_cls, rpc_cb)

# add Project Manager RPC API
for exposed_rpc in EXPOSED_RPCS:
    RPC_DICT[exposed_rpc.__name__] = (exposed_rpc, exe.manager_request)


# events from clients
EVENT_DICT: ws_server.EVENT_DICT_TYPE = {}


async def aio_main() -> None:

    await asyncio.gather(exe.project_manager_client(handle_manager_incoming_messages), _initialize_server())


def print_openapi_models() -> None:

    modules = [arcor2_execution_data.events, events, runtime_events]

    for _, mod in inspect.getmembers(evts, inspect.ismodule):
        modules.append(mod)

    event_types: list[type[events.Event]] = []

    for mod in modules:
        for _, cls in inspect.getmembers(mod, inspect.isclass):
            if issubclass(cls, events.Event):
                event_types.append(cls)

    print(
        models.generate_openapi(
            "ARCOR2 ARServer", arcor2_arserver.version(), [v[0] for k, v in RPC_DICT.items()], event_types
        )
    )


def main() -> None:

    parser = argparse.ArgumentParser()

    parser.add_argument("-v", "--verbose", help="Increase verbosity.", action="store_const", const=True, default=False)
    parser.add_argument(
        "-d",
        "--debug",
        help="Set logging level to debug.",
        action="store_const",
        const=LogLevel.DEBUG,
        default=LogLevel.DEBUG if env.get_bool("ARCOR2_ARSERVER_DEBUG") else LogLevel.INFO,
    )
    parser.add_argument(
        "--version", action="version", version=arcor2_arserver.version(), help="Shows version and exits."
    )
    parser.add_argument(
        "--api_version", action="version", version=arcor2_arserver_data.version(), help="Shows API version and exits."
    )
    parser.add_argument(
        "-a",
        "--asyncio_debug",
        help="Turn on asyncio debug mode.",
        action="store_const",
        const=True,
        default=env.get_bool("ARCOR2_ARSERVER_ASYNCIO_DEBUG"),
    )
    parser.add_argument("--openapi", action="store_true", help="Prints OpenAPI models and exits.")

    args = parser.parse_args()

    if args.openapi:
        print_openapi_models()
        return

    logger.level = args.debug
    glob.VERBOSE = args.verbose

    loop = asyncio.get_event_loop()
    loop.set_debug(enabled=args.asyncio_debug)
    loop.set_exception_handler(ws_server.custom_exception_handler)

    compile_json_schemas()

    run(aio_main(), loop=loop)

    shutil.rmtree(settings.OBJECT_TYPE_PATH)


if __name__ == "__main__":
    main()



from typing import AsyncIterator

from arcor2.cached import CachedProject, UpdateableCachedProject
from arcor2.data import common
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver.clients import project_service as storage


async def scene_names() -> set[str]:
    return {scn.name for scn in (await storage.get_scenes())}


async def project_names() -> set[str]:
    return {proj.name for proj in (await storage.get_projects())}


async def associated_projects(scene_id: str) -> set[str]:
    return {project.id async for project in projects(scene_id)}


async def remove_object_references_from_projects(obj_id: str) -> None:

    assert glob.LOCK.scene

    updated_project_ids: set[str] = set()

    async for project in projects_using_object_as_parent(glob.LOCK.scene.id, obj_id):

        # action_ids: set[str] = set()

        # delete actions using the object
        for action_to_delete in {act.id for act in project.actions if act.parse_type()[0] == obj_id}:
            project.remove_action(action_to_delete)

        # delete actions using obj's action points as parameters
        # TODO fix this!
        """
        actions_using_invalid_param: set[str] = \
            {act.id for act in ap.actions for param in act.parameters
             if param.type in (ActionParameterTypeEnum.JOINTS, ActionParameterTypeEnum.POSE) and
             param.value.startswith(obj_id)}

        ap.actions = [act for act in ap.actions if act.id not in actions_using_invalid_param]

        # get IDs of remaining actions
        action_ids.update({act.id for act in ap.actions})
        """

        # valid_ids: set[str] = action_ids | ActionIOEnum.set()

        # TODO remove invalid logic items

        await storage.update_project(project)
        updated_project_ids.add(project.id)

    logger.info("Updated projects: {}".format(updated_project_ids))


async def projects(scene_id: str) -> AsyncIterator[UpdateableCachedProject]:

    for project_meta in await storage.get_projects():

        project = await storage.get_project(project_meta.id)

        if project.scene_id != scene_id:
            continue

        yield UpdateableCachedProject(project)


def _project_using_object_as_parent(project: CachedProject, obj_id: str) -> bool:

    for ap in project.action_points:
        if ap.parent == obj_id:
            return True

    return False


def _project_referencing_object(project: CachedProject, obj_id: str) -> bool:

    for action in project.actions:
        action_obj_id, _ = action.parse_type()

        if action_obj_id == obj_id:
            return True

    return False


async def projects_using_object(scene_id: str, obj_id: str) -> AsyncIterator[UpdateableCachedProject]:
    """Combines functionality of projects_using_object_as_parent and
    projects_referencing_object.

    :param scene_id:
    :param obj_id:
    :return:
    """

    async for project in projects(scene_id):
        if _project_using_object_as_parent(project, obj_id) or _project_referencing_object(project, obj_id):
            yield project


async def projects_using_object_as_parent(scene_id: str, obj_id: str) -> AsyncIterator[UpdateableCachedProject]:

    async for project in projects(scene_id):
        if _project_using_object_as_parent(project, obj_id):
            yield project


async def invalidate_joints_using_object_as_parent(obj: common.SceneObject) -> None:

    assert glob.LOCK.scene

    # Invalidates robot joints if action point's parent has changed its pose.
    async for project in projects_using_object_as_parent(glob.LOCK.scene.id, obj.id):

        for ap in project.action_points:

            if ap.parent != obj.id:
                continue

            logger.debug(f"Invalidating joints for {project.name}/{ap.name}.")
            project.invalidate_joints(ap.id)

        await storage.update_project(project)


async def projects_referencing_object(scene_id: str, obj_id: str) -> AsyncIterator[CachedProject]:

    async for project in projects(scene_id):
        if _project_referencing_object(project, obj_id):
            yield project



from contextlib import asynccontextmanager
from typing import AsyncGenerator

from arcor2.exceptions import Arcor2Exception
from arcor2_arserver import globals as glob
from arcor2_arserver.decorators import retry
from arcor2_arserver.lock.common import ObjIds, obj_ids_to_list
from arcor2_arserver.lock.exceptions import CannotLock, LockingException


def unique_name(name: str, existing_names: set[str]) -> None:

    if not name:
        raise Arcor2Exception("Name has to be set.")

    if name in existing_names:
        raise Arcor2Exception("Name already exists.")


def make_name_unique(orig_name: str, names: set[str]) -> str:

    cnt = 1
    name = orig_name

    while name in names:
        name = f"{orig_name}_{cnt}"
        cnt += 1

    return name


@asynccontextmanager
async def ctx_write_lock(
    obj_ids: ObjIds, owner: str, auto_unlock: bool = True, dry_run: bool = False
) -> AsyncGenerator[None, None]:
    @retry(exc=CannotLock, tries=glob.LOCK._lock_retries, delay=glob.LOCK._retry_wait)
    async def lock():
        if not await glob.LOCK.write_lock(obj_ids, owner):
            raise CannotLock(glob.LOCK.ErrMessages.LOCK_FAIL.value)

    await lock()
    try:
        yield
    except Arcor2Exception:
        if not dry_run and not auto_unlock:
            await glob.LOCK.write_unlock(obj_ids, owner)
        raise
    finally:
        if dry_run or auto_unlock:
            await glob.LOCK.write_unlock(obj_ids, owner)


@asynccontextmanager
async def ctx_read_lock(
    obj_ids: ObjIds, owner: str, auto_unlock: bool = True, dry_run: bool = False
) -> AsyncGenerator[None, None]:
    """Acquires and releases read lock for objects, unless they are already
    locked.

    :param obj_ids:
    :param owner:
    :param auto_unlock:
    :param dry_run:
    :return:
    """

    @retry(exc=CannotLock, tries=glob.LOCK._lock_retries, delay=glob.LOCK._retry_wait)
    async def lock() -> None:

        if not await glob.LOCK.read_lock(obj_ids, owner):
            raise CannotLock(glob.LOCK.ErrMessages.LOCK_FAIL.value)

    already_locked = False

    obj_ids = obj_ids_to_list(obj_ids)

    for obj_id in obj_ids:
        if not (await glob.LOCK.is_read_locked(obj_id, owner) or await glob.LOCK.is_write_locked(obj_id, owner)):
            await lock()
            break
    else:
        already_locked = True

    try:
        yield
    except Arcor2Exception:
        if not dry_run and not auto_unlock and not already_locked:
            await glob.LOCK.read_unlock(obj_ids, owner)
        raise
    finally:
        if (dry_run or auto_unlock) and not already_locked:
            await glob.LOCK.read_unlock(obj_ids, owner)


async def ensure_write_locked(obj_id: str, owner: str, locked_tree: bool = False) -> None:
    """Check if object is write locked."""

    if not await glob.LOCK.is_write_locked(obj_id, owner, locked_tree):
        raise LockingException("Object is not write locked.")


async def ensure_read_locked(obj_id: str, owner: str, locked_tree: bool = False) -> None:
    """Check if object is read locked."""

    if not await glob.LOCK.is_read_locked(obj_id, owner):
        raise LockingException("Object is not write locked.")


async def get_unlocked_objects(obj_ids: ObjIds, owner: str) -> set[str]:
    """Check if objects are write locked.

    :return: list of objects that are not write locked
    """

    return {obj_id for obj_id in obj_ids_to_list(obj_ids) if not await glob.LOCK.is_write_locked(obj_id, owner)}



import asyncio
from dataclasses import dataclass, field
from typing import NamedTuple

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2.cached import CachedProject, CachedScene
from arcor2.clients import aio_scene_service as scene_srv
from arcor2.data import events, rpc
from arcor2.data.common import Parameter, Pose, Position, SceneObject
from arcor2.data.object_type import Model3dType
from arcor2.data.scene import MeshFocusAction
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import CollisionObject, GenericWithPose, Robot
from arcor2.source.utils import tree_to_str
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver import settings
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.helpers import ctx_write_lock, ensure_write_locked
from arcor2_arserver.object_types.data import ObjectTypeData
from arcor2_arserver.object_types.source import new_object_type
from arcor2_arserver.object_types.utils import add_ancestor_actions, object_actions, remove_object_type
from arcor2_arserver.objects_actions import update_object_model
from arcor2_arserver.robot import check_eef_arm, get_end_effector_pose
from arcor2_arserver.scene import (
    can_modify_scene,
    ensure_scene_started,
    get_instance,
    get_robot_instance,
    scenes,
    update_scene_object_pose,
)
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data import rpc as srpc


@dataclass
class AimedObject:

    obj_id: str
    robot: rpc.common.RobotArg
    poses: dict[int, Pose] = field(default_factory=dict)


_objects_being_aimed: dict[str, AimedObject] = {}  # key == user_name


async def object_aiming_start_cb(req: srpc.o.ObjectAimingStart.Request, ui: WsClient) -> None:
    """Starts the aiming process for a selected object (with mesh) and robot.

    Only possible when the scene is started/online.
    UI have to acquire write locks for object and robot in advance.
    :param req:
    :param ui:
    :return:
    """

    scene = glob.LOCK.scene_or_exception()

    if glob.LOCK.project:
        raise Arcor2Exception("Project has to be closed first.")

    ensure_scene_started()

    user_name = glob.USERS.user_name(ui)

    if user_name in _objects_being_aimed:
        raise Arcor2Exception("Aiming already started.")

    obj_id = req.args.object_id
    scene_obj = scene.object(obj_id)

    obj_type = glob.OBJECT_TYPES[scene_obj.type].meta

    if not obj_type.has_pose:
        raise Arcor2Exception("Only available for objects with pose.")

    if not obj_type.object_model or obj_type.object_model.type != Model3dType.MESH:
        raise Arcor2Exception("Only available for objects with mesh model.")

    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    if not focus_points:
        raise Arcor2Exception("focusPoints not defined for the mesh.")

    await ensure_write_locked(req.args.object_id, user_name)
    await ensure_write_locked(req.args.robot.robot_id, user_name)

    await check_eef_arm(get_robot_instance(req.args.robot.robot_id), req.args.robot.arm_id, req.args.robot.end_effector)

    if req.dry_run:
        return

    _objects_being_aimed[user_name] = AimedObject(req.args.object_id, req.args.robot)
    logger.info(
        f"{user_name} just started aiming of {scene_obj.name} using {scene.object(req.args.robot.robot_id).name}."
    )


class AimingTuple(NamedTuple):
    obj: AimedObject
    user_name: str


async def object_aiming_prune() -> None:
    """Deletes records for users that already lost their locks.

    :return:
    """

    to_delete: list[str] = []

    # users in db but not holding a lock for the object should be deleted
    for un, fo in _objects_being_aimed.items():

        if not await glob.LOCK.is_write_locked(fo.obj_id, un):
            logger.info(f"Object aiming cancelled for {un}.")
            to_delete.append(un)

    for td in to_delete:
        _objects_being_aimed.pop(td, None)


async def object_aiming_check(ui: WsClient) -> AimingTuple:
    """Gets object that is being aimed by the user or exception.

    :param ui:
    :return:
    """

    user_name = glob.USERS.user_name(ui)

    try:
        fo = _objects_being_aimed[user_name]
    except KeyError:
        raise Arcor2Exception("Aiming has to be started first.")

    await ensure_write_locked(fo.obj_id, user_name)
    await ensure_write_locked(fo.robot.robot_id, user_name)

    return AimingTuple(fo, user_name)


async def object_aiming_cancel_cb(req: srpc.o.ObjectAimingCancel.Request, ui: WsClient) -> None:
    """Cancel aiming of the object.

    :param req:
    :param ui:
    :return:
    """

    fo, user_name = await object_aiming_check(ui)

    if req.dry_run:
        return

    _objects_being_aimed.pop(user_name, None)
    if glob.LOCK.scene:
        logger.info(f"Aiming for {glob.LOCK.scene.object(fo.obj_id).name} cancelled by {user_name}.")


async def object_aiming_add_point_cb(
    req: srpc.o.ObjectAimingAddPoint.Request, ui: WsClient
) -> srpc.o.ObjectAimingAddPoint.Response:

    scene = glob.LOCK.scene_or_exception()
    fo, user_name = await object_aiming_check(ui)

    pt_idx = req.args.point_idx
    scene_obj = scene.object(fo.obj_id)
    obj_type = glob.OBJECT_TYPES[scene_obj.type].meta

    assert obj_type.has_pose
    assert obj_type.object_model
    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if pt_idx < 0 or pt_idx > len(focus_points) - 1:
        raise Arcor2Exception("Index out of range.")

    robot_id, end_effector, arm_id = fo.robot.as_tuple()

    robot_inst = get_robot_instance(robot_id)

    r = srpc.o.ObjectAimingAddPoint.Response()
    r.data = r.Data(finished_indexes=list(fo.poses.keys()))

    if not req.dry_run:
        fo.poses[pt_idx] = await get_end_effector_pose(robot_inst, end_effector, arm_id)
        r.data = r.Data(finished_indexes=list(fo.poses.keys()))
        logger.info(
            f"{user_name} just aimed index {pt_idx} for {scene_obj.name}. Done indexes: {r.data.finished_indexes}."
        )

    return r


async def object_aiming_done_cb(req: srpc.o.ObjectAimingDone.Request, ui: WsClient) -> None:
    """Calls scene service to get a new pose for the object.

    In case of success, robot and object are kept locked, unlocking is responsibility of ui.
    On failure, UI may do another attempt or call ObjectAimingCancel.

    :param req:
    :param ui:
    :return:
    """

    scene = glob.LOCK.scene_or_exception()
    fo, user_name = await object_aiming_check(ui)

    obj_type = glob.OBJECT_TYPES[scene.object(fo.obj_id).type].meta
    assert obj_type.object_model
    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if len(fo.poses) < len(focus_points):
        raise Arcor2Exception(f"Only {len(fo.poses)} points were done out of {len(focus_points)}.")

    obj = scene.object(fo.obj_id)
    assert obj.pose

    obj_inst = get_instance(fo.obj_id, CollisionObject)

    if req.dry_run:
        return

    fp: list[Position] = []
    rp: list[Position] = []

    for idx, pose in fo.poses.items():

        fp.append(focus_points[idx].position)
        rp.append(pose.position)

    mfa = MeshFocusAction(fp, rp)

    logger.debug(f"Attempt to aim object {obj_inst.name}, data: {mfa}")

    try:
        new_pose = await scene_srv.focus(mfa)  # TODO how long does it take?
    except scene_srv.SceneServiceException as e:
        logger.error(f"Aiming failed with: {e}, mfa: {mfa}.")
        raise Arcor2Exception(f"Aiming failed. {str(e)}") from e

    logger.info(f"Done aiming for {obj_inst.name}.")

    _objects_being_aimed.pop(user_name, None)
    asyncio.create_task(update_scene_object_pose(scene, obj, new_pose, obj_inst))
    return None


async def new_object_type_cb(req: srpc.o.NewObjectType.Request, ui: WsClient) -> None:

    async with ctx_write_lock(glob.LOCK.SpecialValues.ADDING_OBJECT, glob.USERS.user_name(ui)):
        meta = req.args

        if meta.type in glob.OBJECT_TYPES:
            raise Arcor2Exception("Object type already exists.")

        hlp.is_valid_type(meta.type)

        if meta.base not in glob.OBJECT_TYPES:
            raise Arcor2Exception(
                f"Unknown base object type '{meta.base}', " f"known types are: {', '.join(glob.OBJECT_TYPES.keys())}."
            )

        base = glob.OBJECT_TYPES[meta.base]

        if base.meta.disabled:
            raise Arcor2Exception("Base object is disabled.")

        assert base.type_def is not None

        if issubclass(base.type_def, Robot):
            raise Arcor2Exception("Can't subclass Robot.")

        meta.has_pose = issubclass(base.type_def, GenericWithPose)

        if issubclass(base.type_def, CollisionObject):
            if not meta.object_model:
                raise Arcor2Exception("Objects based on CollisionObject must have collision model.")
        else:
            if meta.object_model:
                raise Arcor2Exception("Only objects based on CollisionObject can have collision model.")

        if req.dry_run:
            return None

        obj = meta.to_object_type()
        ast = new_object_type(glob.OBJECT_TYPES[meta.base].meta, meta)
        obj.source = tree_to_str(ast)

        if meta.object_model:
            await update_object_model(meta, meta.object_model)

        type_def = await hlp.run_in_executor(
            hlp.save_and_import_type_def,
            obj.source,
            obj.id,
            base.type_def,
            settings.OBJECT_TYPE_PATH,
            settings.OBJECT_TYPE_MODULE,
        )
        assert issubclass(type_def, base.type_def)
        actions = object_actions(type_def, ast)

        meta.modified = await storage.update_object_type(obj)

        glob.OBJECT_TYPES[meta.type] = ObjectTypeData(meta, type_def, actions, ast)
        add_ancestor_actions(meta.type, glob.OBJECT_TYPES)

        evt = sevts.o.ChangedObjectTypes([meta])
        evt.change_type = events.Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_object_model_cb(req: srpc.o.UpdateObjectModel.Request, ui: WsClient) -> None:

    can_modify_scene()
    glob.LOCK.scene_or_exception(True)  # only allow while editing scene

    obj_data = glob.OBJECT_TYPES[req.args.object_type_id]

    if not obj_data.type_def:
        raise Arcor2Exception("ObjectType disabled.")

    if not issubclass(obj_data.type_def, CollisionObject):
        raise Arcor2Exception("Not a CollisionObject.")

    assert obj_data.meta.object_model
    assert obj_data.ast

    if req.args.object_model == obj_data.meta.object_model:
        raise Arcor2Exception("No change requested.")

    await ensure_write_locked(req.args.object_type_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    await update_object_model(obj_data.meta, req.args.object_model)
    obj_data.meta.object_model = req.args.object_model

    ot = obj_data.meta.to_object_type()
    ot.source = tree_to_str(obj_data.ast)

    obj_data.meta.modified = await storage.update_object_type(ot)

    evt = sevts.o.ChangedObjectTypes([obj_data.meta])
    evt.change_type = events.Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))


async def get_object_actions_cb(req: srpc.o.GetActions.Request, ui: WsClient) -> srpc.o.GetActions.Response:
    return srpc.o.GetActions.Response(data=list(glob.OBJECT_TYPES[req.args.type].actions.values()))


async def get_object_types_cb(req: srpc.o.GetObjectTypes.Request, ui: WsClient) -> srpc.o.GetObjectTypes.Response:
    return srpc.o.GetObjectTypes.Response(data=[obj.meta for obj in glob.OBJECT_TYPES.values()])


def check_scene_for_object_type(scene: CachedScene, object_type: str) -> None:

    if object_type in scene.object_types:
        raise Arcor2Exception(f"Used in scene {scene.name}.")


async def delete_object_type_cb(
    req: srpc.o.DeleteObjectTypes.Request, ui: WsClient
) -> srpc.o.DeleteObjectTypes.Response:
    async def _delete_model(obj_type: ObjectTypeData) -> None:

        # do not care so much if delete_model fails
        if not obj_type.meta.object_model:
            return

        try:
            await storage.delete_model(obj_type.meta.object_model.model().id)
        except storage.ProjectServiceException as e:
            logger.error(str(e))

    async def _delete_ot(ot: str) -> None:

        obj_type = glob.OBJECT_TYPES[ot]

        if obj_type.meta.built_in:
            raise Arcor2Exception("Can't delete built-in type.")

        for obj in glob.OBJECT_TYPES.values():
            if obj.meta.base == ot:
                raise Arcor2Exception(f"Object type is base of '{obj.meta.type}'.")

        async for scene in scenes():
            check_scene_for_object_type(scene, ot)

        if glob.LOCK.scene:
            check_scene_for_object_type(glob.LOCK.scene, ot)

        if req.dry_run:
            return

        await asyncio.gather(storage.delete_object_type(ot), _delete_model(obj_type), remove_object_type(ot))

        await glob.LOCK.write_unlock(ot, user_name)  # need to be unlocked while it exists in glob.OBJECT_TYPES
        del glob.OBJECT_TYPES[ot]

        evt = sevts.o.ChangedObjectTypes([obj_type.meta])
        evt.change_type = events.Event.Type.REMOVE
        asyncio.create_task(notif.broadcast_event(evt))

    user_name = glob.USERS.user_name(ui)

    obj_types_to_delete: list[str] = (
        list(req.args)
        if req.args is not None
        else [obj.meta.type for obj in glob.OBJECT_TYPES.values() if not obj.meta.built_in]
    )

    response = srpc.o.DeleteObjectTypes.Response()
    response.data = []

    async with ctx_write_lock(obj_types_to_delete, user_name, auto_unlock=False, dry_run=req.dry_run):

        res = await asyncio.gather(*[_delete_ot(ot) for ot in obj_types_to_delete], return_exceptions=True)

        for idx, r in enumerate(res):
            if isinstance(r, Arcor2Exception):
                response.data.append(srpc.o.DeleteObjectTypes.Response.Data(obj_types_to_delete[idx], str(r)))
            else:
                assert r is None

    if not response.data:
        response.data = None

    if response.data:
        response.result = False
        response.messages = []
        response.messages.append("Failed to delete one or more ObjectTypes.")

    return response


def check_override(
    scene: CachedScene, project: CachedProject, obj_id: str, override: Parameter, add_new_one: bool = False
) -> SceneObject:

    obj = scene.object(obj_id)

    for par in glob.OBJECT_TYPES[obj.type].meta.settings:
        if par.name == override.name:
            if par.type != override.type:
                raise Arcor2Exception("Override can't change parameter type.")
            break
    else:
        raise Arcor2Exception("Unknown parameter name.")

    if add_new_one:
        try:
            for existing_override in project.overrides[obj.id]:
                if override.name == existing_override.name:
                    raise Arcor2Exception("Override already exists.")
        except KeyError:
            pass
    else:
        if obj.id not in project.overrides:
            raise Arcor2Exception("There are no overrides for the object.")

        for override in project.overrides[obj.id]:
            if override.name == override.name:
                break
        else:
            raise Arcor2Exception("Override not found.")

    return obj


async def add_override_cb(req: srpc.o.AddOverride.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    project = glob.LOCK.project_or_exception()

    obj = check_override(scene, project, req.args.id, req.args.override, add_new_one=True)

    await ensure_write_locked(req.args.id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    if obj.id not in project.overrides:
        project.overrides[obj.id] = []

    project.overrides[obj.id].append(req.args.override)
    project.update_modified()

    evt = sevts.o.OverrideUpdated(req.args.override)
    evt.change_type = events.Event.Type.ADD
    evt.parent_id = req.args.id
    asyncio.ensure_future(notif.broadcast_event(evt))


async def update_override_cb(req: srpc.o.UpdateOverride.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    project = glob.LOCK.project_or_exception()

    obj = check_override(scene, project, req.args.id, req.args.override)

    await ensure_write_locked(req.args.id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    for override in project.overrides[obj.id]:
        if override.name == override.name:
            override.value = req.args.override.value
    project.update_modified()

    evt = sevts.o.OverrideUpdated(req.args.override)
    evt.change_type = events.Event.Type.UPDATE
    evt.parent_id = req.args.id
    asyncio.ensure_future(notif.broadcast_event(evt))


async def delete_override_cb(req: srpc.o.DeleteOverride.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    project = glob.LOCK.project_or_exception()

    obj = check_override(scene, project, req.args.id, req.args.override)

    await ensure_write_locked(req.args.id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    project.overrides[obj.id] = [ov for ov in project.overrides[obj.id] if ov.name != req.args.override.name]

    if not project.overrides[obj.id]:
        del project.overrides[obj.id]

    project.update_modified()

    evt = sevts.o.OverrideUpdated(req.args.override)
    evt.change_type = events.Event.Type.REMOVE
    evt.parent_id = req.args.id
    asyncio.ensure_future(notif.broadcast_event(evt))


async def object_type_usage_cb(req: srpc.o.ObjectTypeUsage.Request, ui: WsClient) -> srpc.o.ObjectTypeUsage.Response:

    resp = srpc.o.ObjectTypeUsage.Response()
    resp.data = set()  # mypy does not recognize it correctly with Response(data=set())

    async for scene in scenes():
        if req.args.id in scene.object_types:
            resp.data.add(scene.id)

    return resp



import asyncio
import base64
import os
from typing import TYPE_CHECKING

import aiofiles
import websockets
from aiofiles import tempfile
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2 import rest
from arcor2.data import common, rpc
from arcor2.data.rpc import get_id
from arcor2.exceptions import Arcor2Exception
from arcor2_arserver import events as server_events
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver import project
from arcor2_arserver.scene import scene_started, start_scene, stop_scene
from arcor2_arserver_data import events as sevts
from arcor2_build_data import URL as BUILD_URL
from arcor2_execution_data import URL as EXE_URL
from arcor2_execution_data import rpc as erpc

if TYPE_CHECKING:
    ReqQueue = asyncio.Queue[rpc.common.RPC.Request]
    RespQueue = asyncio.Queue[rpc.common.RPC.Response]
else:
    ReqQueue = asyncio.Queue
    RespQueue = asyncio.Queue

MANAGER_RPC_REQUEST_QUEUE: ReqQueue = ReqQueue()
MANAGER_RPC_RESPONSES: dict[int, RespQueue] = {}


async def run_temp_package(package_id: str, start_paused: bool = False, breakpoints: None | set[str] = None) -> None:

    # TODO lock scene and project?

    assert glob.LOCK.scene
    assert glob.LOCK.project
    project_id = glob.LOCK.project.id
    glob.TEMPORARY_PACKAGE = True

    scene_online = scene_started()

    if scene_online:
        await stop_scene(glob.LOCK.scene)  # the package will start it on its own

    await project.close_project(show_mainscreen_after_that=False)
    req = erpc.RunPackage.Request
    exe_req = req(get_id(), args=req.Args(package_id, start_paused, breakpoints))
    exe_resp = await manager_request(exe_req)

    if not exe_resp.result:
        logger.warning(f"Execution of temporary package failed with: {exe_resp.messages}.")
    else:
        await server_events.package_started.wait()
        await server_events.package_stopped.wait()
        logger.info("Temporary package stopped, let's remove it and reopen project.")

    glob.TEMPORARY_PACKAGE = False

    await manager_request(erpc.DeletePackage.Request(get_id(), args=rpc.common.IdArgs(package_id)))

    await project.open_project(project_id)

    assert glob.LOCK.scene
    assert glob.LOCK.project

    await notif.broadcast_event(
        sevts.p.OpenProject(sevts.p.OpenProject.Data(glob.LOCK.scene.scene, glob.LOCK.project.project))
    )

    if scene_online:
        await start_scene(glob.LOCK.scene)


async def build_and_upload_package(project_id: str, package_name: str) -> str:
    """Builds package and uploads it to the Execution unit.

    :param project_id:
    :param package_name:
    :return: generated package ID.
    """

    package_id = common.uid("pkg")

    # call build service
    # TODO store data in memory
    async with tempfile.TemporaryDirectory() as tmpdirname:
        path = os.path.join(tmpdirname, "publish.zip")

        await hlp.run_in_executor(
            rest.download,
            f"{BUILD_URL}/project/publish",
            path,
            {
                "packageName": package_name,
                "projectId": project_id,
            },
        )

        async with aiofiles.open(path, "rb") as zip_file:
            b64_bytes = base64.b64encode(await zip_file.read())
            b64_str = b64_bytes.decode()

    # send data to execution service
    exe_req = erpc.UploadPackage.Request(get_id(), args=erpc.UploadPackage.Request.Args(package_id, b64_str))
    exe_resp = await manager_request(exe_req)

    if not exe_resp.result:
        if not exe_resp.messages:
            raise Arcor2Exception("Upload to the Execution unit failed.")
        raise Arcor2Exception("\n".join(exe_resp.messages))

    return package_id


async def manager_request(req: rpc.common.RPC.Request, ui: None | WsClient = None) -> rpc.common.RPC.Response:

    assert req.id not in MANAGER_RPC_RESPONSES

    MANAGER_RPC_RESPONSES[req.id] = RespQueue(maxsize=1)
    await MANAGER_RPC_REQUEST_QUEUE.put(req)
    resp = await MANAGER_RPC_RESPONSES[req.id].get()
    del MANAGER_RPC_RESPONSES[req.id]
    return resp


async def project_manager_client(handle_manager_incoming_messages) -> None:

    while True:

        logger.info("Attempting connection to manager...")

        try:

            async with websockets.connect(EXE_URL) as manager_client:  # type: ignore  # TODO not sure what is wrong

                logger.info("Connected to manager.")

                future = asyncio.ensure_future(handle_manager_incoming_messages(manager_client))

                while True:

                    if future.done():
                        break

                    try:
                        msg = await asyncio.wait_for(MANAGER_RPC_REQUEST_QUEUE.get(), 1.0)
                    except asyncio.TimeoutError:
                        continue

                    try:
                        await manager_client.send(msg.to_json())
                    except websockets.exceptions.ConnectionClosed:
                        await MANAGER_RPC_REQUEST_QUEUE.put(msg)
                        break
        except ConnectionRefusedError as e:
            logger.error(e)
            await asyncio.sleep(delay=1.0)



import asyncio
import math
import time
from typing import Awaitable, Callable

import numpy as np
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import env
from arcor2 import transformations as tr
from arcor2 import ws_server
from arcor2.clients.project_service import URL as ps_url
from arcor2.data import common
from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import run_in_executor
from arcor2.object_types.abstract import Camera, MultiArmRobot, Robot
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver import robot
from arcor2_arserver.helpers import ctx_read_lock, ctx_write_lock, ensure_write_locked
from arcor2_arserver.scene import (
    ensure_scene_started,
    get_instance,
    get_robot_instance,
    scene_started,
    update_scene_object_pose,
)
from arcor2_arserver_data import rpc as srpc
from arcor2_arserver_data.events.common import ProcessState
from arcor2_arserver_data.events.robot import HandTeachingMode
from arcor2_calibration_data import client as calib_client
from arcor2_calibration_data.client import CalibrateRobotArgs
from arcor2_runtime.events import RobotEef, RobotJoints

RBT_CALIB = "RobotCalibration"

TaskDict = dict[str, asyncio.Task]

ROBOT_JOINTS_TASKS: TaskDict = {}
EEF_POSE_TASKS: TaskDict = {}

EVENT_PERIOD = env.get_float("ARCOR2_STREAMING_PERIOD", 0.1)

if EVENT_PERIOD <= 0:
    EVENT_PERIOD = 0.1


async def robot_joints_event(robot_inst: Robot) -> None:

    logger.info(f"Sending joints for {robot_inst.name} started.")
    while scene_started() and glob.ROBOT_JOINTS_REGISTERED_UIS[robot_inst.id]:

        start = time.monotonic()

        try:
            evt = RobotJoints(RobotJoints.Data(robot_inst.id, (await robot.get_robot_joints(robot_inst, None, True))))
        except Arcor2Exception as e:
            logger.error(f"Failed to get joints for {robot_inst.name}. {str(e)}")
            await asyncio.sleep(1)
            continue

        evt_json = evt.to_json()
        await asyncio.gather(
            *[ws_server.send_json_to_client(ui, evt_json) for ui in glob.ROBOT_JOINTS_REGISTERED_UIS[robot_inst.id]]
        )

        end = time.monotonic()
        await asyncio.sleep(EVENT_PERIOD - (end - start))

    ROBOT_JOINTS_TASKS.pop(robot_inst.id, None)

    # TODO notify UIs that registration was cancelled
    glob.ROBOT_JOINTS_REGISTERED_UIS.pop(robot_inst.id, None)

    logger.info(f"Sending joints for {robot_inst.name} stopped.")


async def eef_pose(robot_inst: Robot, eef_id: str, arm_id: None | str = None) -> RobotEef.Data.EefPose:

    return RobotEef.Data.EefPose(eef_id, (await robot.get_end_effector_pose(robot_inst, eef_id, arm_id)), arm_id)


async def robot_eef_pose_event(robot_inst: Robot) -> None:

    eefs: dict[None | str, set[str]] = {}

    try:
        try:
            for arm_id in await robot.get_arms(robot_inst):
                eefs[arm_id] = await robot.get_end_effectors(robot_inst, arm_id)
        except robot.SingleArmRobotException:
            eefs = {None: await robot.get_end_effectors(robot_inst, None)}
    except Arcor2Exception as e:
        logger.error(f"Failed to start sending poses for {robot_inst.name}. {str(e)}")
    else:

        logger.info(f"Sending poses for {robot_inst.name} started. Arms/EEFs: {eefs}.")

        while scene_started() and glob.ROBOT_EEF_REGISTERED_UIS[robot_inst.id]:

            start = time.monotonic()

            evt = RobotEef(RobotEef.Data(robot_inst.id))

            try:
                evt.data.end_effectors = await asyncio.gather(
                    *[eef_pose(robot_inst, eef_id, arm_id) for arm_id in eefs.keys() for eef_id in eefs[arm_id]]
                )
            except Arcor2Exception as e:
                logger.error(f"Failed to get eef pose for {robot_inst.name}. {str(e)}")
                await asyncio.sleep(1)
                continue

            evt_json = evt.to_json()
            await asyncio.gather(
                *[ws_server.send_json_to_client(ui, evt_json) for ui in glob.ROBOT_EEF_REGISTERED_UIS[robot_inst.id]]
            )

            end = time.monotonic()
            await asyncio.sleep(EVENT_PERIOD - (end - start))

    EEF_POSE_TASKS.pop(robot_inst.id, None)

    # TODO notify UIs that registration was cancelled
    glob.ROBOT_EEF_REGISTERED_UIS.pop(robot_inst.id, None)

    logger.info(f"Sending poses for {robot_inst.name} stopped.")


async def get_robot_meta_cb(req: srpc.r.GetRobotMeta.Request, ui: WsClient) -> srpc.r.GetRobotMeta.Response:

    return srpc.r.GetRobotMeta.Response(
        data=[obj.robot_meta for obj in glob.OBJECT_TYPES.values() if obj.robot_meta is not None]
    )


async def get_robot_joints_cb(req: srpc.r.GetRobotJoints.Request, ui: WsClient) -> srpc.r.GetRobotJoints.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetRobotJoints.Response(
            data=await robot.get_robot_joints(get_robot_instance(req.args.robot_id), req.args.arm_id)
        )


async def get_end_effector_pose_cb(
    req: srpc.r.GetEndEffectorPose.Request, ui: WsClient
) -> srpc.r.GetEndEffectorPose.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetEndEffectorPose.Response(
            data=await robot.get_end_effector_pose(
                get_robot_instance(req.args.robot_id), req.args.end_effector_id, req.args.arm_id
            )
        )


async def get_end_effectors_cb(req: srpc.r.GetEndEffectors.Request, ui: WsClient) -> srpc.r.GetEndEffectors.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetEndEffectors.Response(
            data=await robot.get_end_effectors(get_robot_instance(req.args.robot_id), req.args.arm_id)
        )


async def get_robot_arms_cb(req: srpc.r.GetRobotArms.Request, ui: WsClient) -> srpc.r.GetRobotArms.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetRobotArms.Response(data=await robot.get_arms(get_robot_instance(req.args.robot_id)))


async def get_grippers_cb(req: srpc.r.GetGrippers.Request, ui: WsClient) -> srpc.r.GetGrippers.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetGrippers.Response(
            data=await robot.get_grippers(get_robot_instance(req.args.robot_id), req.args.arm_id)
        )


async def get_suctions_cb(req: srpc.r.GetSuctions.Request, ui: WsClient) -> srpc.r.GetSuctions.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetSuctions.Response(
            data=await robot.get_suctions(get_robot_instance(req.args.robot_id), req.args.arm_id)
        )


async def register(
    req: srpc.r.RegisterForRobotEvent.Request,
    robot_inst: Robot,
    ui: WsClient,
    tasks: TaskDict,
    reg_uis: glob.RegisteredUiDict,
    coro: Callable[[Robot], Awaitable[None]],
) -> None:

    if req.args.send:

        reg_uis[req.args.robot_id].add(ui)

        if req.args.robot_id not in tasks:
            # start task
            tasks[req.args.robot_id] = asyncio.create_task(coro(robot_inst))  # type: ignore   # TODO figure it out

    else:
        try:
            reg_uis[req.args.robot_id].remove(ui)
        except KeyError as e:
            raise Arcor2Exception("Failed to unregister.") from e

        # cancel task if not needed anymore
        if not reg_uis[req.args.robot_id]:
            task = tasks[req.args.robot_id]

            if not task.cancelled():
                task.cancel()

            del tasks[req.args.robot_id]


async def register_for_robot_event_cb(req: srpc.r.RegisterForRobotEvent.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()

        # check if robot exists
        robot_inst = get_robot_instance(req.args.robot_id)

        if req.args.what == req.args.RegisterEnum.JOINTS:
            await register(
                req, robot_inst, ui, ROBOT_JOINTS_TASKS, glob.ROBOT_JOINTS_REGISTERED_UIS, robot_joints_event
            )
        elif req.args.what == req.args.RegisterEnum.EEF_POSE:

            if isinstance(robot_inst, MultiArmRobot):
                for arm_id in await robot.get_arms(robot_inst):
                    if await robot.get_end_effectors(robot_inst, arm_id):
                        break
                else:
                    raise Arcor2Exception("Robot does not have any end effector.")

            elif not await robot.get_end_effectors(robot_inst, None):
                raise Arcor2Exception("Robot does not have any end effector.")

            await register(req, robot_inst, ui, EEF_POSE_TASKS, glob.ROBOT_EEF_REGISTERED_UIS, robot_eef_pose_event)
        else:
            raise Arcor2Exception(f"Option '{req.args.what.value}' not implemented.")

        return None


async def check_feature(robot_inst: Robot, feature_name: str) -> None:

    obj_type = glob.OBJECT_TYPES[robot_inst.__class__.__name__]

    if obj_type.robot_meta is None:
        raise Arcor2Exception("Not a robot.")

    try:
        if not getattr(obj_type.robot_meta.features, feature_name):
            raise Arcor2Exception(f"Robot does not support '{feature_name}' feature.")
    except AttributeError:
        raise Arcor2Exception(f"Unknown robot feature: {feature_name}.")


async def move_to_pose_cb(req: srpc.r.MoveToPose.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()
    user_name = glob.USERS.user_name(ui)

    async with ctx_write_lock(req.args.robot_id, user_name, auto_unlock=False):
        ensure_scene_started()

        robot_inst = get_robot_instance(req.args.robot_id)
        await robot.check_eef_arm(robot_inst, req.args.arm_id, req.args.end_effector_id)

        await check_feature(robot_inst, Robot.move_to_pose.__name__)
        await robot.check_robot_before_move(robot_inst)

        if (req.args.position is None) != (req.args.orientation is None):

            target_pose = await robot.get_end_effector_pose(robot_inst, req.args.end_effector_id, req.args.arm_id)

            if req.args.position:
                target_pose.position = req.args.position
            elif req.args.orientation:
                target_pose.orientation = req.args.orientation

        elif req.args.position is not None and req.args.orientation is not None:
            target_pose = common.Pose(req.args.position, req.args.orientation)
        else:
            raise Arcor2Exception("Position or orientation should be given.")

        # TODO check if the target pose is reachable (dry_run)
        asyncio.ensure_future(
            robot.move_to_pose(
                robot_inst,
                req.args.end_effector_id,
                req.args.arm_id,
                target_pose,
                req.args.speed,
                req.args.safe,
                req.args.linear,
                user_name,
            )
        )


async def move_to_joints_cb(req: srpc.r.MoveToJoints.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()
    user_name = glob.USERS.user_name(ui)

    async with ctx_write_lock(req.args.robot_id, user_name, auto_unlock=False):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot_id)
        await check_feature(robot_inst, Robot.move_to_joints.__name__)
        await robot.check_robot_before_move(robot_inst)

        asyncio.ensure_future(
            robot.move_to_joints(robot_inst, req.args.joints, req.args.speed, req.args.safe, user_name, req.args.arm_id)
        )


async def stop_robot_cb(req: srpc.r.StopRobot.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()

    # Stop robot cannot use lock, because robot is locked when action is called. Stop will also release lock.
    ensure_scene_started()
    robot_inst = get_robot_instance(req.args.robot_id)
    await check_feature(robot_inst, Robot.stop.__name__)
    await robot.stop(robot_inst)


async def move_to_action_point_cb(req: srpc.r.MoveToActionPoint.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    project = glob.LOCK.project_or_exception()

    async with ctx_write_lock(req.args.robot_id, glob.USERS.user_name(ui)):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot_id)

        await robot.check_robot_before_move(robot_inst)

        if (req.args.orientation_id is None) == (req.args.joints_id is None):
            raise Arcor2Exception("Set orientation or joints. Not both.")

        if req.args.orientation_id:

            await check_feature(robot_inst, Robot.move_to_pose.__name__)

            if req.args.end_effector_id is None:
                raise Arcor2Exception("eef id has to be set.")

            pose = tr.abs_pose_from_ap_orientation(scene, project, req.args.orientation_id)

            # TODO check if the target pose is reachable (dry_run)
            asyncio.ensure_future(
                robot.move_to_ap_orientation(
                    robot_inst,
                    req.args.end_effector_id,
                    req.args.arm_id,
                    pose,
                    req.args.speed,
                    req.args.orientation_id,
                    req.args.safe,
                    req.args.linear,
                )
            )

        elif req.args.joints_id:

            await check_feature(robot_inst, Robot.move_to_joints.__name__)

            # TODO add arm_id to ProjectRobotJoints and use it as a parameter for move_to_ap_joints
            joints = project.joints(req.args.joints_id)

            # TODO check if the joints are within limits and reachable (dry_run)
            asyncio.ensure_future(
                robot.move_to_ap_joints(
                    robot_inst, joints.joints, req.args.speed, req.args.joints_id, req.args.safe, req.args.arm_id
                )
            )


async def ik_cb(req: srpc.r.InverseKinematics.Request, ui: WsClient) -> srpc.r.InverseKinematics.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock([req.args.robot_id, req.args.end_effector_id], glob.USERS.user_name(ui)):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot_id)
        await check_feature(robot_inst, Robot.inverse_kinematics.__name__)

        joints = await robot.ik(
            robot_inst,
            req.args.end_effector_id,
            req.args.arm_id,
            req.args.pose,
            req.args.start_joints,
            req.args.avoid_collisions,
        )
        resp = srpc.r.InverseKinematics.Response()
        resp.data = joints
        return resp


async def fk_cb(req: srpc.r.ForwardKinematics.Request, ui: WsClient) -> srpc.r.ForwardKinematics.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock([req.args.robot_id, req.args.end_effector_id], glob.USERS.user_name(ui)):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot_id)
        await check_feature(robot_inst, Robot.forward_kinematics.__name__)

        pose = await robot.fk(robot_inst, req.args.end_effector_id, req.args.arm_id, req.args.joints)
        resp = srpc.r.ForwardKinematics.Response()
        resp.data = pose
        return resp


async def calibrate_robot(
    robot_inst: Robot, camera_inst: Camera, move_to_calibration_pose: bool, user_name: str
) -> None:

    try:
        scene = glob.LOCK.scene_or_exception()

        assert camera_inst.color_camera_params

        await notif.broadcast_event(ProcessState(ProcessState.Data(RBT_CALIB, ProcessState.Data.StateEnum.Started)))

        try:

            if move_to_calibration_pose:
                await run_in_executor(robot_inst.move_to_calibration_pose)
            robot_joints = await run_in_executor(robot_inst.robot_joints)
            depth_image = await run_in_executor(camera_inst.depth_image, 128)

            args = CalibrateRobotArgs(
                robot_joints,
                robot_inst.pose,
                camera_inst.pose,
                camera_inst.color_camera_params,
                f"{ps_url}/models/{robot_inst.urdf_package_name}/mesh/file",
            )

            new_pose = await run_in_executor(calib_client.calibrate_robot, args, depth_image)

        except Arcor2Exception as e:
            await notif.broadcast_event(
                ProcessState(ProcessState.Data(RBT_CALIB, ProcessState.Data.StateEnum.Failed, str(e)))
            )
            logger.exception("Failed to calibrate the robot.")
            return

        await update_scene_object_pose(scene, scene.object(robot_inst.id), new_pose, robot_inst)
        await notif.broadcast_event(ProcessState(ProcessState.Data(RBT_CALIB, ProcessState.Data.StateEnum.Finished)))
    finally:
        await glob.LOCK.write_unlock(camera_inst.id, user_name)


async def calibrate_robot_cb(req: srpc.r.CalibrateRobot.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()
    user_name = glob.USERS.user_name(ui)

    ensure_scene_started()

    if not req.args.camera_id:
        for obj in glob.SCENE_OBJECT_INSTANCES.values():
            if isinstance(obj, Camera):
                req.args.camera_id = obj.id
                break
        else:
            raise Arcor2Exception("No camera found.")

    async with ctx_write_lock(req.args.camera_id, user_name, auto_unlock=False):

        robot_inst = get_robot_instance(req.args.robot_id)

        if not robot_inst.urdf_package_name:
            raise Arcor2Exception("Robot with model required!")

        camera_inst = get_instance(req.args.camera_id, Camera)

        if camera_inst.color_camera_params is None:
            raise Arcor2Exception("Calibrated camera required!")

        await ensure_write_locked(req.args.robot_id, user_name)

        asyncio.ensure_future(calibrate_robot(robot_inst, camera_inst, req.args.move_to_calibration_pose, user_name))

        return None


async def hand_teaching_mode_cb(req: srpc.r.HandTeachingMode.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()

    ensure_scene_started()
    robot_inst = get_robot_instance(req.args.robot_id)

    # in this case, method name does not correspond to feature name
    await check_feature(robot_inst, "hand_teaching")

    if robot_inst.move_in_progress:
        raise Arcor2Exception("Not possible while robot moves.")

    params = []
    if isinstance(robot_inst, MultiArmRobot):
        params.append(req.args.arm_id)
    elif req.args.arm_id:
        raise Arcor2Exception(
            f"The '{req.args.arm_id}' arm_id was set for single arm robot '{robot_inst.name}' "
            f"of type '{robot_inst.__class__.__name__}'!"
        )

    hand_teaching_mode = await run_in_executor(robot_inst.get_hand_teaching_mode, *params)

    if req.args.enable == hand_teaching_mode:
        raise Arcor2Exception("That's the current state.")

    await ensure_write_locked(req.args.robot_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    await run_in_executor(robot_inst.set_hand_teaching_mode, req.args.enable, *params)
    evt = HandTeachingMode(HandTeachingMode.Data(req.args.robot_id, req.args.enable, req.args.arm_id))
    asyncio.ensure_future(notif.broadcast_event(evt))


async def step_robot_eef_cb(req: srpc.r.StepRobotEef.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()

    ensure_scene_started()
    robot_inst = get_robot_instance(req.args.robot_id)

    await check_feature(robot_inst, Robot.move_to_pose.__name__)
    await robot.check_robot_before_move(robot_inst)

    tp = await robot.get_end_effector_pose(robot_inst, req.args.end_effector_id, req.args.arm_id)

    if req.args.mode == req.args.mode.ROBOT:
        tp = tr.make_pose_rel(robot_inst.pose, tp)
    elif req.args.mode == req.args.mode.RELATIVE:
        assert req.args.pose
        tp = tr.make_pose_rel(req.args.pose, tp)
    elif req.args.mode == req.args.mode.USER:
        assert req.args.pose
        raise Arcor2Exception("Not supported yet.")

    if req.args.what == req.args.what.POSITION:
        if req.args.axis == req.args.axis.X:
            tp.position.x += req.args.step
        elif req.args.axis == req.args.axis.Y:
            tp.position.y += req.args.step
        elif req.args.axis == req.args.axis.Z:
            tp.position.z += req.args.step
    elif req.args.what == req.args.what.ORIENTATION:
        if req.args.axis == req.args.axis.X:
            tp.orientation *= common.Orientation.from_rotation_vector(x=req.args.step)
        elif req.args.axis == req.args.axis.Y:
            tp.orientation *= common.Orientation.from_rotation_vector(y=req.args.step)
        elif req.args.axis == req.args.axis.Z:
            tp.orientation *= common.Orientation.from_rotation_vector(z=req.args.step)

    if req.args.mode == req.args.mode.ROBOT:
        tp = tr.make_pose_abs(robot_inst.pose, tp)
    elif req.args.mode == req.args.mode.RELATIVE:
        assert req.args.pose
        tp = tr.make_pose_abs(req.args.pose, tp)

    await robot.check_reachability(scene, robot_inst, req.args.end_effector_id, req.args.arm_id, tp, req.args.safe)

    await ensure_write_locked(req.args.robot_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    asyncio.ensure_future(
        robot.move_to_pose(
            robot_inst, req.args.end_effector_id, req.args.arm_id, tp, req.args.speed, req.args.safe, req.args.linear
        )
    )


async def set_eef_perpendicular_to_world_cb(req: srpc.r.SetEefPerpendicularToWorld.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()

    ensure_scene_started()
    robot_inst = get_robot_instance(req.args.robot_id)

    await check_feature(robot_inst, Robot.move_to_pose.__name__)
    await check_feature(robot_inst, Robot.inverse_kinematics.__name__)
    await check_feature(robot_inst, Robot.forward_kinematics.__name__)
    await robot.check_robot_before_move(robot_inst)

    await ensure_write_locked(req.args.robot_id, glob.USERS.user_name(ui))

    if req.dry_run:  # attempt to find suitable joints can take some time so it is not done for dry_run
        return

    tp, current_joints = await robot.get_pose_and_joints(robot_inst, req.args.end_effector_id, req.args.arm_id)

    target_joints_diff: float = 0.0
    winning_idx: int = -1

    poses: list[common.Pose] = [
        common.Pose(
            tp.position,
            common.Orientation.from_rotation_vector(y=math.pi) * common.Orientation.from_rotation_vector(z=z_rot),
        )
        for z_rot in np.linspace(-math.pi, math.pi, 360)
    ]

    # select best (closest joint configuration) reachable pose
    tasks = [
        robot.ik(robot_inst, req.args.end_effector_id, req.args.arm_id, pose, current_joints, req.args.safe)
        for pose in poses
    ]

    # order of results from gather corresponds to order of tasks
    for idx, res in enumerate(await asyncio.gather(*tasks, return_exceptions=True)):

        if not isinstance(res, list):
            continue

        diff = 0.0

        for f, b in zip(current_joints, res):
            assert (
                f.name == b.name
            ), f"current joints: {[j.name for j in current_joints]}, joints from ik {[j.name for j in res]}."
            diff += (f.value - b.value) ** 2

        if winning_idx < 0:
            target_joints_diff = diff
            winning_idx = idx
        elif diff < target_joints_diff:
            winning_idx = idx
            target_joints_diff = diff

    if winning_idx < 0:
        raise Arcor2Exception("Could not find reachable pose.")

    asyncio.ensure_future(
        robot.move_to_pose(
            robot_inst,
            req.args.end_effector_id,
            req.args.arm_id,
            poses[winning_idx],
            req.args.speed,
            req.args.safe,
            req.args.linear,
        )
    )




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/python/arcor2_arserver:arcor2_arserver_dist

from setuptools import setup

setup(**{
    'author': 'Robo@FIT',
    'author_email': 'imaterna@fit.vut.cz',
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.10',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
    ],
    'description': 'ARCOR2 ARServer',
    'entry_points': {
        'console_scripts': [
            'arcor2_arserver = arcor2_arserver.scripts.arserver:main',
            'arcor2_broadcaster = arcor2_arserver.scripts.broadcaster:main',
        ],
    },
    'install_requires': (
        'aiofiles~=22.1.0',
        'aiologger~=0.7.0',
        'aiorun~=2022.11.1',
        'apispec-webframeworks~=0.5.2',
        'apispec[yaml]~=6.0.2',
        'arcor2_arserver_data~=1.0.0',
        'arcor2_build_data~=1.0.0',
        'arcor2_calibration_data~=1.0.0',
        'arcor2_execution_data~=1.0.0',
        'arcor2_runtime~=1.0.0',
        'arcor2~=1.0.0',
        'dataclasses-jsonschema[apispec,fast-dateparsing,fast-uuid,fast-validation]~=2.16.0',
        'lru-dict~=1.1.8',
        'numpy-quaternion[numba,scipy]~=2022.4.2',
        'numpy~=1.23.5',
        'pyhumps==3.8.0',
        'typing-inspect~=0.8.0',
        'websockets~=10.4',
    ),
    'license': 'LGPL',
    'long_description': """# arcor2_arserver

## Environment variables

- `ARCOR2_ARSERVER_PORT=6789` - by default, the service will listen on port 6789.
- `ARCOR2_STREAMING_PERIOD=0.1` - controls the period of streaming a robot's EEF poses and joints. 

### Caching

- `ARCOR2_ARSERVER_CACHE_TIMEOUT=1.0` - by default, ARServer checks for changes in scenes, projects or ObjectTypes max. once per second.
- `ARCOR2_ARSERVER_CACHE_SCENES=32` - by default, ARServer keeps 32 last used scenes in its cache.
- `ARCOR2_ARSERVER_CACHE_PROJECTS=64` - by default, ARServer keeps 64 last used projects in its cache.
- `ARCOR2_ARSERVER_CACHE_OBJECT_TYPES=32` - by default, ARServer keeps 64 last used ObjectTypes in its cache.

### Resources locking

- `ARCOR2_ARSERVER_LOCK_TIMEOUT=300` - if a user do not log again within this timeout (in seconds), their locks are discarded.
  - Can be used to workaround 'Locking failed' issues - if set to some low value, it is sufficient to log out and after the locks got discarded, log in again.
- `ARCOR2_ARSERVER_LOCK_RETRIES=13` - number of retries when trying to acquire the global lock.
- `ARCOR2_ARSERVER_LOCK_RETRY_WAIT=0.15` - controls how long to wait between attempts to get the global lock. 

### Debugging

- `ARCOR2_MAX_RPC_DURATION=0.1` - by default, a warning is emitted when any RPC call takes longer than 0.1 second.
- `ARCOR2_ARSERVER_DEBUG=1` - switches logger to the `DEBUG` level. 
- `ARCOR2_ARSERVER_ASYNCIO_DEBUG=1` - turns on `asyncio` debug output (helpful to debug problems related to concurrency). 
- `ARCOR2_REST_DEBUG=1` - may be used to debug problems related to communication with the Project, Scene Build and Calibration services.
# Changelog

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),

## [1.0.0] - 2023-02-14

### Changed

- Dependency on arcor2 1.0.0.
- Compatibility with the newest version of the Project service.
- ARServer now also talks to Asset service (checks whether mesh/URDF exists).

## [0.27.0] - 2022-12-12

### Changed

- Dependency on `arcor2~=0.26.0`.

## [0.26.0] - 2022-10-28

### Changed

- Switched to Python 3.10, updated dependencies.

## [0.25.0] - 2022-08-17

### Changed

- Adaptation to changes in API of the Build service.
- Updated scene client (changed API).

## [0.24.0] - 2022-04-25

### Added

- New RPCs `GetScene` and `GetProject`.

## [0.23.0] - 2022-01-25

### Changed

- Breaking change of WebSockets API (`ActionStateBefore` event).
- Switched to Python 3.9, updated dependencies.

## [0.22.1] - 2021-10-26

### Changed

- Added HTTP upload server for logs from AREditor
  - Logs are uploaded to /root/logs (could be changed using ARCOR2_AREDITOR_LOGS_FOLDER env variable)

## [0.22.0] - 2021-10-25

### Changed

- When getting pose/joints from a robot fails, continue trying instead of giving up.
- Handle `KinematicsException` to deliver a more precise error message.
- Updating object models.
  - New RPC `UpdateObjectModel`.
  - Ability to lock individual ObjectTypes.
  - Various improvements regarding locking.
- Proper handling of scene/project problems.
  - Any problem makes scene/project invalid, and then it can't be opened.  
  - Scene/project problems are now checked when:
    - Scene/project is listed.
    - Opening a scene/project is requested.
  - Found problems are cached for better performance.
  - ObjectTypes are checked for changes during each listing of scenes/projects.
- Deleting ObjectTypes reworked.
  - `DeleteObjectType` replaced with `DeleteObjectTypes`.
  - The new RPC takes list of ObjectType IDs.
  - Empty list means all ObjectTypes.
  - It returns list of id/error for types that could not be removed.
  - This is especially useful for usage with `dry_run`.
- Updated environment variables, please see README.
- Improved logging to debug 'Locking failed'.
- Support for virtual collision objects.
  - `AddVirtualCollisionObjectToScene` RPC creates a new OT and adds its instance to a scene.
  - In the simplest case, each collision object will have its own ObjectType.
  - But it is also possible to add more instances as with other types.
  - When the last instance of VCO is removed, ARServer also removes the ObjectType automatically.
- Ignoring `hidden` actions (a new flag in `ActionMetadata`).
- New RPC `ObjectTypeUsage` returning scene ids where the ObjectType is used.
- Support for debugging - relaying related RPCs to the Execution Service.
- Better handling of exceptions in asyncio tasks.
  - More useful traceback is printed out.
  - In development mode, an unhandled exception leads to shutdown of a service.
  - In production mode, traceback is printed out and the service tries to continue.

### Fixed

- Ban HT mode changes while moving.
- Arguments for getting robot joints (only affected multi-arm robots).
- Handle locking exceptions on scene start/stop.
- Copying of action points.
  - Fixed locking within `CopyActionPoint`.
  - Fixed update of orientation IDs.
- Added locking to Project client to make the cache caching properly.
- `modified`  was missing for disabled object.
  - This happened when it was not possible to get object's base.
- Copy AP - update of orientation ID in actions.
- Removed 'delayed' unlocking of resources which caused race conditions in some cases and made the crashes hard to debug.
- HT mode for single arm robots was broken.


## [0.21.0] - 2021-08-05

### Changed

- Locks are not automatically released when aiming is done or cancelled.
- Linear movements are used by default for stepping and table alignment.
- Joints stream (`RobotJoints` event) now contains also gripper joints when available (depends on the particular ObjectType).

### Fixed

- Getting action results (the result was always `False`).
- `ctx_read_lock` made some RPCs impossible to success.
- Object parameters checking.
  - Values and types of object parameters were not checked at all.
  - For now, only basic types are supported.

## [0.20.0] - 2021-07-29

### Changed

- Support for multi arm robots.
  - Dealing with `arm_id` parameter in RPCs.
  - Streaming EEF poses for all arms.
- Allow to reposition scene objects while online.
  - It was allowed to reposition objects using robot,
  - ...but disallowed to do it manually.
  - This was inconsistent and unnecessarily restrictive.
  - Now objects (except robots) can be repositioned anytime.
- When creating a new project, common parameters are added automatically.
  - At the moment those are `scene_id` and `project_id`.
  - A user could use or remove them.
- Collisions are not removed after stopping the scene.
- ObjectType change detection is now based on `modified` instead of comparing source codes.
- ARServer Project service client now stores cached scenes/projects.
- Object aiming reworked.
  - RPCs were renamed and all of them now have `dry_run`.
  - There is a new method to cancel the process.
  - One user can be aiming one object at the moment.
  - Both the object and the selected robot have to be write-locked.
  - Improved logging.
  - Test was added.
- Server no longer downloads URDF packages.
  - `ARCOR2_DATA_PATH` environment variable is no longer used.
  - Clients will get URDFs through reverse proxy from Project service.
- `NewObjectType` no longer assumes that mesh already exists.
  - A file associated to mesh must exist in advance.
  
  
### Fixed
- All calls to the Scene service are guarded against exceptions.
- Renaming of project parameters fixed.
  - It was possible to use already taken name.

## [0.19.0] - 2021-06-22

- Update to API version 0.15.0.

## [0.18.0] - 2021-06-11

### Changed

- Updated to work with Project service 0.8.0.
- Allow fast re-login after connection was broken.
- Added locking to const-related RPCs.
- Log calibration-related errors.

### Fixed

- Handle invalid ObjectType settings annotation.

## [0.17.0] - 2021-05-21

### Changed

- Return value of `GetCameraPose` RPC updated.
- Objects initialization order.
  - Object initialization order can be set using class-level `INIT_PRIORITY` variable.
  - The higher priority, the sooner are objects of that type initialized.
  - Initialization happens in parallel (for objects of the same priority).
- `SetEefPerpendicularToWorld` now calls IK in parallel.
  - This provides almost 100% speed-up even when the robot service proceeds requests one by one.
- Added locking of internal resources.
  - Created a lock module that allows to read/write lock existing or special objects. The module contains:
    - lock object with all necessary methods;
    - structures for keeping data about locked objects;
    - locking-related exceptions;
    - queue and task for notifying UI about lock data;
    - methods for auto-unlocking after timeout when user logouts.
  - Base tests of locking structure.
  - RPCs for (un)locking object, registering user name.
  - Events for (un)locking object.
  - Created a class for maintaining connected UIs.
  - Project methods to get object by ID, object parent and object children.
  
  - All RPCs that requires some kind of locking are now lock-guarded.
  - Updated existing tests to work with newly implemented locking.
  - Global variables `SCENE`, `PROJECT` and `INTERFACES` moved to new classes.
- Rewritten cache for the Project service.
  - The cache is now always on (provides a significant speed-up).
  - External updates to the Project service are recognized properly.
- New RPC `AddApUsingRobot`.
- Support for multiple inheritance.
    - ObjectTypes can now use mixins.
    - It should be used like `class NewObjectType(MixinA, MixinB, Generic)`.
    - E.g. the last ancestor should be something derived from `Generic`.
- `SetEefPerpendicularToWorld` faster as it calls IK in parallel. 

### Fixed

- `SceneState` event was not sent to a newly connected UIs.
- Results of actions are now forgotten when going offline.
- Fixed support for `link` parameter type.

## [0.16.0] - 2021-04-20

### Fixed
- Action can't use its own result as a parameter (link parameter type).

### Changed
- Dependency on arcor2 0.15.0 with updated REST client.
- Action parameter value is now always JSON.
  - It used to be a string for link/constant special types.


## [0.15.1] - 2021-04-19

### Fixed
- Dependency on bugfix release of arcor2.

## [0.15.0] - 2021-03-30

### Changed
- New RPC `SetEefPerpendicularToWorld`.
- New RPC `StepRobotEef`.

### Fixed
- `CopyActionPoint` RPC: update action parameters to use copied orientations.
- Scene now goes offline before execution and to the original state after it.
- Scene-related messages adapted to a new terminology.

## [0.14.0] - 2021-03-15

### Changed
- Improved logging for debugging robot features.
- Suffix is added to copied AP/action instead of prefix.

### Fixed
- Names for orientations/joints were not checked as valid identifiers.
- Action names were checked, but should not be.

## [0.13.0] - 2021-03-03

### Changed

-  API to copy action point (including orientations, joints, child APs).
- Robot-related API (MoveToPose, MoveToJoints, MoveToActionPoint) updated in order to contain `safe` parameter.
- API to enable/disable robot hand teaching mode.  
- Concrete and helpful error messages when a name entered by the user is refused.

### Fixed
- Only `BareActionPoint` was sent on parent change.

## [0.12.0] - 2021-02-08

### Changed
- Improved scene-related messages.
  - Error messages now contain the name of the object which caused the error.
- New `MarkersCorners` RPC for getting markers corners.
- Support for `an` parameter (explicit parameters).
- Improved check of action signature and annotation.
- New WS API for execution state reporting.
- Support for project constants.
  - Only basic types are supported so far.

### Fixed
- Better handling of missing collision model.
  - If server can't get collision model of the object, the object is disabled.
  - Previously, this situation caused crash of the server.
- `GetCameraPose` callback fixed (used to be `Calibration`).

## [0.11.0] - 2020-12-14

### Changed
- Improved support for logic branching.
  - So far, only boolean outputs are supported.
- Improved handling of return values (uses the same function as the main script).
- The internal database of ObjectTypes in updated each time before opening a scene or a project.
  - Changes are notified using `ChangedObjectTypes` event (one event for added objects, one for updated and another one for deleted ones).
- Support for camera/robot calibration.
- Initial support for cameras (getting images, camera parameters).

## [0.10.1] - 2020-10-30

### Changed
- RPC for IK/FK.

## [0.10.0] - 2020-10-22

### Changed
- Updated due to a breaking change in `arcor2_execution_data`.

## [0.9.1] - 2020-10-19

### Changed
- ARCOR2 dependency updated

### Fixed
- fixed check for existing override
 
## [0.9.0] - 2020-10-16

### Changed
- WS API for updates of scene objects parameters and management of project-defined overrides.
- Integration tests refactored into multiple modules that can run in parallel.
- `UpdateObjectPose` and `UpdateActionPointPosition` now has dry_run.
- Box/Cylinder/Sphere models now have some constraints on their dimensions.
- 

## [0.8.0] - 2020-09-24
### Changed
- The first release of the separated package.
- RPC/Events refactored. 
- ARServer script now has ```--openapi``` argument and is able to generate OpenAPI models for RPCs and events.
- Detection of loops in: AP hierarchy, logic.
- Caching client for Project service.
- Proper handling of object inheritance.
- ObjectTypes are stored in temporary module on disk.
- Integration tests.""",
    'long_description_content_type': 'text/markdown',
    'name': 'arcor2_arserver',
    'namespace_packages': (
    ),
    'package_data': {
        'arcor2_arserver': (
            'VERSION',
            'py.typed',
        ),
    },
    'packages': (
        'arcor2_arserver',
        'arcor2_arserver.clients',
        'arcor2_arserver.lock',
        'arcor2_arserver.object_types',
        'arcor2_arserver.rpc',
        'arcor2_arserver.scripts',
    ),
    'python_requires': '==3.10.*',
    'version': '1.0.0',
})



import asyncio
import copy
import functools
from contextlib import asynccontextmanager
from typing import AsyncGenerator

import quaternion
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2.cached import UpdateableCachedScene
from arcor2.data import common, object_type
from arcor2.data.events import Event, PackageState
from arcor2.exceptions import Arcor2Exception
from arcor2.image import image_from_str
from arcor2.object_types.abstract import Generic, VirtualCollisionObject
from arcor2.source.utils import tree_to_str
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver import settings
from arcor2_arserver.checks import check_object_parameters
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.common import (
    associated_projects,
    project_names,
    projects_using_object,
    remove_object_references_from_projects,
)
from arcor2_arserver.helpers import (
    ctx_read_lock,
    ctx_write_lock,
    ensure_write_locked,
    get_unlocked_objects,
    unique_name,
)
from arcor2_arserver.lock.exceptions import LockingException
from arcor2_arserver.object_types.data import ObjectTypeData
from arcor2_arserver.object_types.source import new_object_type
from arcor2_arserver.object_types.utils import add_ancestor_actions, object_actions
from arcor2_arserver.objects_actions import get_object_types, update_object_model
from arcor2_arserver.robot import check_eef_arm, get_end_effector_pose
from arcor2_arserver.scene import (
    add_object_to_scene,
    can_modify_scene,
    clear_auto_remove_schedule,
    delete_if_not_used,
    ensure_scene_started,
    get_instance,
    get_robot_instance,
    get_scene_problems,
    get_scene_state,
    notify_scene_closed,
    notify_scene_opened,
    open_scene,
    save_scene,
    scene_names,
    scene_started,
    scenes,
    schedule_auto_remove,
    start_scene,
    stop_scene,
    unschedule_auto_remove,
    update_scene_object_pose,
)
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data import rpc as srpc
from arcor2_arserver_data.objects import ObjectTypeMeta
from arcor2_calibration_data import client as calibration


@asynccontextmanager
async def managed_scene(scene_id: str, make_copy: bool = False) -> AsyncGenerator[UpdateableCachedScene, None]:

    save_back = False

    if glob.LOCK.scene and glob.LOCK.scene.id == scene_id:
        if make_copy:
            scene = copy.deepcopy(glob.LOCK.scene)
            save_back = True
        else:
            scene = glob.LOCK.scene
    else:
        save_back = True
        scene = UpdateableCachedScene(await storage.get_scene(scene_id))

    if make_copy:
        scene.id = common.Scene.uid()

    try:
        yield scene
    finally:
        if save_back:
            asyncio.ensure_future(storage.update_scene(scene))


async def new_scene_cb(req: srpc.s.NewScene.Request, ui: WsClient) -> None:
    """Creates and opens a new scene on the server. Fails if any scene is open
    or if scene id/name already exists.

    :param req:
    :return:
    """

    async with glob.LOCK.get_lock(dry_run=req.dry_run):
        if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
            raise Arcor2Exception("Can't create scene while package runs.")

        if glob.LOCK.scene:
            raise Arcor2Exception("Scene has to be closed first.")

        unique_name(req.args.name, (await scene_names()))

        # TODO workaround for https://gitlab.com/kinalisoft/test-it-off/project/-/issues/16
        unique_name(req.args.name, (await project_names()))

        if req.dry_run:
            return None

        # TODO get_object_types - not ideal, may take quite long time
        #  ...maybe creating a scene should be done similarly to starting one (process with events about its state)?
        await get_object_types()

        glob.LOCK.scene = UpdateableCachedScene(common.Scene(req.args.name, description=req.args.description))
        asyncio.ensure_future(notify_scene_opened(sevts.s.OpenScene(sevts.s.OpenScene.Data(glob.LOCK.scene.scene))))

        return None


async def close_scene_cb(req: srpc.s.CloseScene.Request, ui: WsClient) -> None:
    """Closes scene on the server.

    :param req:
    :return:
    """

    async with ctx_write_lock(glob.LOCK.SpecialValues.SCENE, glob.USERS.user_name(ui), dry_run=req.dry_run):

        scene = glob.LOCK.scene_or_exception()

        if glob.LOCK.project:
            raise Arcor2Exception("Project has to be closed first.")

        if not req.args.force and scene.has_changes:
            raise Arcor2Exception("Scene has unsaved changes.")

        can_modify_scene()  # can't close scene while started

        if await glob.LOCK.get_locked_roots_count() > 1:
            raise LockingException(glob.LOCK.ErrMessages.SOMETHING_LOCKED.value)

        if req.dry_run:
            return None

        scene_id = scene.id
        glob.LOCK.scene = None
        glob.OBJECTS_WITH_UPDATED_POSE.clear()
        await clear_auto_remove_schedule()
        asyncio.ensure_future(notify_scene_closed(scene_id))


async def save_scene_cb(req: srpc.s.SaveScene.Request, ui: WsClient) -> None:

    async with glob.LOCK.get_lock(req.dry_run):

        scene = glob.LOCK.scene_or_exception()

        if glob.LOCK.project:
            raise Arcor2Exception("Project has to be closed first.")

        if req.dry_run:
            return None

        await save_scene(scene)
        return None


async def open_scene_cb(req: srpc.s.OpenScene.Request, ui: WsClient) -> None:

    async with glob.LOCK.get_lock():
        if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
            raise Arcor2Exception("Can't open scene while package runs.")

        if glob.LOCK.scene:
            raise Arcor2Exception("Scene already opened.")

        await open_scene(req.args.id)

        assert glob.LOCK.scene
        assert not glob.LOCK.scene.has_changes
        asyncio.ensure_future(
            notify_scene_opened(sevts.s.OpenScene(data=sevts.s.OpenScene.Data(glob.LOCK.scene.scene)))
        )
        return None


async def list_scenes_cb(req: srpc.s.ListScenes.Request, ui: WsClient) -> srpc.s.ListScenes.Response:

    resp = srpc.s.ListScenes.Response()
    resp.data = []

    async for scene in scenes():
        assert scene.created
        assert scene.modified
        resp.data.append(
            resp.Data(
                scene.id, scene.name, scene.created, scene.modified, scene.description, await get_scene_problems(scene)
            )
        )

    return resp


async def get_scene_cb(req: srpc.s.GetScene.Request, ui: WsClient) -> srpc.s.GetScene.Response:

    return srpc.s.GetScene.Response(data=(await storage.get_scene(req.args.id)).scene)


async def add_object_to_scene_cb(req: srpc.s.AddObjectToScene.Request, ui: WsClient) -> None:

    async with ctx_write_lock(glob.LOCK.SpecialValues.SCENE, glob.USERS.user_name(ui)):

        scene = glob.LOCK.scene_or_exception()

        if glob.LOCK.project:
            raise Arcor2Exception("Project has to be closed first.")

        can_modify_scene()

        obj = common.SceneObject(req.args.name, req.args.type, req.args.pose, req.args.parameters)

        await add_object_to_scene(scene, obj, dry_run=req.dry_run)

        if req.dry_run:
            return None

        scene.update_modified()

        asyncio.create_task(unschedule_auto_remove(obj.type))  # TODO could this even be case?

        evt = sevts.s.SceneObjectChanged(obj)
        evt.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_object_parameters_cb(req: srpc.s.UpdateObjectParameters.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)

    can_modify_scene()

    obj = scene.object(req.args.id)

    if obj.type not in glob.OBJECT_TYPES:
        raise Arcor2Exception("Unknown object type.")

    obj_type = glob.OBJECT_TYPES[obj.type]

    check_object_parameters(obj_type, req.args.parameters)

    await ensure_write_locked(req.args.id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    obj.parameters = req.args.parameters
    scene.update_modified()

    evt = sevts.s.SceneObjectChanged(obj)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def scene_object_usage_request_cb(
    req: srpc.s.SceneObjectUsage.Request, ui: WsClient
) -> srpc.s.SceneObjectUsage.Response:
    """Works for both services and objects.

    :param req:
    :return:
    """

    scene = glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.id, glob.USERS.user_name(ui)):
        if not (any(obj.id == req.args.id for obj in scene.objects)):
            raise Arcor2Exception("Unknown ID.")

        resp = srpc.s.SceneObjectUsage.Response()
        resp.data = set()

        async for project in projects_using_object(scene.id, req.args.id):
            resp.data.add(project.id)

        return resp


# TODO move to objects
async def action_param_values_cb(
    req: srpc.o.ActionParamValues.Request, ui: WsClient
) -> srpc.o.ActionParamValues.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.id, glob.USERS.user_name(ui)):
        ensure_scene_started()

        inst = get_instance(req.args.id, Generic)

        parent_params = {}

        for pp in req.args.parent_params:
            parent_params[pp.id] = pp.value

        try:
            method_name, required_parent_params = inst.DYNAMIC_PARAMS[req.args.param_id]
        except KeyError:
            raise Arcor2Exception("Unknown parameter or values not constrained.")

        if parent_params.keys() != required_parent_params:
            raise Arcor2Exception("Not all required parent params were given.")

        # TODO validate method parameters vs parent_params (check types)?

        resp = srpc.o.ActionParamValues.Response()

        try:
            method = getattr(inst, method_name)
        except AttributeError:
            logger.error(
                f"Unable to get values for parameter {req.args.param_id}, "
                f"object/service {inst.id} has no method named {method_name}."
            )
            raise Arcor2Exception("System error.")

        # TODO update hlp.run_in_executor to support kwargs
        resp.data = await asyncio.get_event_loop().run_in_executor(None, functools.partial(method, **parent_params))
        return resp


async def remove_from_scene_cb(req: srpc.s.RemoveFromScene.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)
    user_name = glob.USERS.user_name(ui)

    to_lock = await get_unlocked_objects(req.args.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):

        can_modify_scene()

        if not req.args.force and {proj.name async for proj in projects_using_object(scene.id, req.args.id)}:
            raise Arcor2Exception("Can't remove object that is used in project(s).")

        if req.dry_run:
            return None

        if req.args.id not in scene.object_ids:
            raise Arcor2Exception("Unknown id.")

        await glob.LOCK.write_unlock(req.args.id, user_name)

        obj = scene.object(req.args.id)
        scene.delete_object(req.args.id)

        if req.args.id in glob.OBJECTS_WITH_UPDATED_POSE:
            glob.OBJECTS_WITH_UPDATED_POSE.remove(req.args.id)

        evt = sevts.s.SceneObjectChanged(obj)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))

        # TODO this should be done after scene is saved
        asyncio.ensure_future(remove_object_references_from_projects(req.args.id))

        if (meta := glob.OBJECT_TYPES[obj.type].meta).base == VirtualCollisionObject.__name__:
            await schedule_auto_remove(meta.type)


async def update_object_pose_using_robot_cb(req: srpc.o.UpdateObjectPoseUsingRobot.Request, ui: WsClient) -> None:
    """Updates object's pose using a pose of the robot's end effector.

    :param req:
    :return:
    """

    if req.args.id == req.args.robot.robot_id:
        raise Arcor2Exception("Robot cannot update its own pose.")

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)
    user_name = glob.USERS.user_name(ui)

    to_lock = await get_unlocked_objects([obj for obj in (req.args.robot.robot_id, req.args.id)], user_name)
    async with ctx_write_lock(to_lock, user_name):
        ensure_scene_started()

        robot_inst = get_robot_instance(req.args.robot.robot_id)
        await check_eef_arm(robot_inst, req.args.robot.arm_id, req.args.robot.end_effector)

        scene_object = scene.object(req.args.id)

        obj_type = glob.OBJECT_TYPES[scene_object.type]

        if not obj_type.meta.has_pose:
            raise Arcor2Exception("Object without pose.")

        object_model = obj_type.meta.object_model

        if object_model:
            collision_model = object_model.model()
            if isinstance(collision_model, object_type.Mesh) and req.args.pivot != req.args.PivotEnum.MIDDLE:
                raise Arcor2Exception("Only middle pivot point is supported for objects with mesh collision model.")
        elif req.args.pivot != req.args.PivotEnum.MIDDLE:
            raise Arcor2Exception("Only middle pivot point is supported for objects without collision model.")

        new_pose = await get_end_effector_pose(robot_inst, req.args.robot.end_effector, req.args.robot.arm_id)

        position_delta = common.Position()

        if object_model:
            collision_model = object_model.model()
            if isinstance(collision_model, object_type.Box):
                if req.args.pivot == req.args.PivotEnum.TOP:
                    position_delta.z -= collision_model.size_z / 2
                elif req.args.pivot == req.args.PivotEnum.BOTTOM:
                    position_delta.z += collision_model.size_z / 2
            elif isinstance(collision_model, object_type.Cylinder):
                if req.args.pivot == req.args.PivotEnum.TOP:
                    position_delta.z -= collision_model.height / 2
                elif req.args.pivot == req.args.PivotEnum.BOTTOM:
                    position_delta.z += collision_model.height / 2
            elif isinstance(collision_model, object_type.Sphere):
                if req.args.pivot == req.args.PivotEnum.TOP:
                    position_delta.z -= collision_model.radius / 2
                elif req.args.pivot == req.args.PivotEnum.BOTTOM:
                    position_delta.z += collision_model.radius / 2

        position_delta = position_delta.rotated(new_pose.orientation)

        assert scene_object.pose

        scene_object.pose.position = new_pose.position - position_delta

        scene_object.pose.orientation.set_from_quaternion(
            new_pose.orientation.as_quaternion() * quaternion.quaternion(0, 1, 0, 0)
        )

        asyncio.ensure_future(update_scene_object_pose(scene, scene_object))
        return None


async def update_object_pose_cb(req: srpc.s.UpdateObjectPose.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)

    if scene_started():
        try:
            get_robot_instance(req.args.object_id)
        except Arcor2Exception:
            pass  # ok, it should not be a robot
        else:
            raise Arcor2Exception("Robot's pose can be only updated offline.")

    obj = scene.object(req.args.object_id)

    if not obj.pose:
        raise Arcor2Exception("Object without pose.")

    await ensure_write_locked(req.args.object_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    asyncio.ensure_future(update_scene_object_pose(scene, obj, req.args.pose))
    return None


async def rename_object_cb(req: srpc.s.RenameObject.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)
    target_obj = scene.object(req.args.id)

    if target_obj.name == req.args.new_name:
        raise Arcor2Exception("Name unchanged")

    for obj_name in scene.object_names():
        if obj_name == req.args.new_name:
            raise Arcor2Exception("Object name already exists.")

    hlp.is_valid_identifier(req.args.new_name)

    user_name = glob.USERS.user_name(ui)
    await ensure_write_locked(req.args.id, user_name)

    if req.dry_run:
        return None

    target_obj.name = req.args.new_name

    scene.update_modified()

    evt = sevts.s.SceneObjectChanged(target_obj)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))

    await glob.LOCK.write_unlock(req.args.id, user_name, True)
    return None


async def rename_scene_cb(req: srpc.s.RenameScene.Request, ui: WsClient) -> None:

    unique_name(req.args.new_name, (await scene_names()))

    # TODO workaround for https://gitlab.com/kinalisoft/test-it-off/project/-/issues/16
    unique_name(req.args.new_name, (await project_names()))

    user_name = glob.USERS.user_name(ui)
    await ensure_write_locked(req.args.id, user_name)

    if req.dry_run:
        return None

    async with managed_scene(req.args.id) as scene:
        scene.name = req.args.new_name

        evt = sevts.s.SceneChanged(scene.bare)
        evt.change_type = Event.Type.UPDATE_BASE
        asyncio.ensure_future(notif.broadcast_event(evt))

    await glob.LOCK.write_unlock(req.args.id, user_name, True)
    return None


async def delete_scene_cb(req: srpc.s.DeleteScene.Request, ui: WsClient) -> None | srpc.s.DeleteScene.Response:

    if glob.LOCK.scene:
        raise Arcor2Exception("Scene has to be closed first.")

    user_name = glob.USERS.user_name(ui)

    async with ctx_write_lock(req.args.id, user_name, auto_unlock=req.dry_run):

        if assoc_projects := await associated_projects(req.args.id):
            resp = srpc.s.DeleteScene.Response(result=False)
            resp.messages = ["Scene has associated projects."]
            resp.data = assoc_projects
            await glob.LOCK.write_unlock(req.args.id, user_name)
            return resp

        if req.dry_run:
            return None

        scene = UpdateableCachedScene(await storage.get_scene(req.args.id))
        await glob.LOCK.write_unlock(req.args.id, user_name)
        await storage.delete_scene(req.args.id)

        for obj_type in scene.object_types:
            if (meta := glob.OBJECT_TYPES[obj_type].meta).base == VirtualCollisionObject.__name__:
                logger.debug(f"VCO {meta.type} will be (probably) auto-removed.")
                asyncio.create_task(delete_if_not_used(meta))

        evt = sevts.s.SceneChanged(scene.bare)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def projects_with_scene_cb(
    req: srpc.s.ProjectsWithScene.Request, ui: WsClient
) -> srpc.s.ProjectsWithScene.Response:

    async with ctx_read_lock(req.args.id, glob.USERS.user_name(ui)):
        resp = srpc.s.ProjectsWithScene.Response()
        resp.data = await associated_projects(req.args.id)
        return resp


async def update_scene_description_cb(req: srpc.s.UpdateSceneDescription.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.scene_id, glob.USERS.user_name(ui)):
        async with managed_scene(req.args.scene_id) as scene:
            scene.description = req.args.new_description
            scene.update_modified()

            evt = sevts.s.SceneChanged(scene.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def copy_scene_cb(req: srpc.s.CopyScene.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.source_id, glob.USERS.user_name(ui)):
        # TODO check if target_name is unique
        async with managed_scene(req.args.source_id, make_copy=True) as scene:
            scene.name = req.args.target_name

            evt = sevts.s.SceneChanged(scene.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))

        return None


# TODO maybe this would better fit into another category of RPCs? Like common/misc?
async def get_camera_pose_cb(req: srpc.c.GetCameraPose.Request, ui: WsClient) -> srpc.c.GetCameraPose.Response:

    try:
        return srpc.c.GetCameraPose.Response(
            data=await hlp.run_in_executor(
                calibration.estimate_camera_pose,
                req.args.camera_parameters,
                image_from_str(req.args.image),
                req.args.inverse,
            )
        )
    except calibration.MarkerNotFound:  # this is ok
        raise
    except calibration.CalibrationException as e:  # this means a serious problem and should be logged
        logger.warn(f"Failed to get camera pose. {str(e)}")
        raise


# TODO maybe this would better fit into another category of RPCs? Like common/misc?
async def marker_corners_cb(req: srpc.c.MarkersCorners.Request, ui: WsClient) -> srpc.c.MarkersCorners.Response:

    return srpc.c.MarkersCorners.Response(
        data=await hlp.run_in_executor(
            calibration.markers_corners, req.args.camera_parameters, image_from_str(req.args.image)
        )
    )


async def start_scene_cb(req: srpc.s.StartScene.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()

    if get_scene_state().data.state != sevts.s.SceneState.Data.StateEnum.Stopped:
        raise Arcor2Exception("Scene not stopped.")

    # online scene can't be modified so we demand that UIs free all their locks first
    # when editing project, changes can be done both online and offline
    if not glob.LOCK.project and await glob.LOCK.get_write_locks_count():
        raise LockingException(glob.LOCK.ErrMessages.SOMETHING_LOCKED.value)

    if await glob.LOCK.is_write_locked(glob.LOCK.SpecialValues.SCENE, glob.LOCK.Owners.SERVER):
        raise Arcor2Exception("Scene locked.")

    if await glob.LOCK.is_write_locked(glob.LOCK.SpecialValues.PROJECT, glob.LOCK.Owners.SERVER):
        raise Arcor2Exception("Project locked.")

    if req.dry_run:
        return

    asyncio.ensure_future(start_scene(scene))


async def stop_scene_cb(req: srpc.s.StopScene.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()

    if get_scene_state().data.state != sevts.s.SceneState.Data.StateEnum.Started:
        raise Arcor2Exception("Scene not started.")

    if await glob.LOCK.is_write_locked(glob.LOCK.SpecialValues.SCENE, glob.LOCK.Owners.SERVER):
        raise Arcor2Exception("Scene locked.")

    if await glob.LOCK.is_write_locked(glob.LOCK.SpecialValues.PROJECT, glob.LOCK.Owners.SERVER):
        raise Arcor2Exception("Project locked.")

    if glob.RUNNING_ACTION:  # TODO acquire lock?
        raise Arcor2Exception("There is a running action.")

    if req.dry_run:
        return

    asyncio.ensure_future(stop_scene(scene))


async def add_virtual_collision_object_to_scene_cb(
    req: srpc.s.AddVirtualCollisionObjectToScene.Request, ui: WsClient
) -> None:

    async with ctx_write_lock(
        (glob.LOCK.SpecialValues.SCENE, glob.LOCK.SpecialValues.ADDING_OBJECT), glob.USERS.user_name(ui)
    ):

        scene = glob.LOCK.scene_or_exception()

        if glob.LOCK.project:
            raise Arcor2Exception("Project has to be closed first.")

        can_modify_scene()

        if req.args.name in glob.OBJECT_TYPES:
            raise Arcor2Exception("ObjectType already exists.")

        if req.args.name in scene.object_names():
            raise Arcor2Exception("Name has to be unique.")

        hlp.is_valid_identifier(req.args.name)

        # post_init of ObjectTypeMeta check if ObjectType.id == model.id
        meta = ObjectTypeMeta(
            req.args.name, base=VirtualCollisionObject.__name__, object_model=req.args.model, has_pose=True
        )

        if req.dry_run:
            return None

        ast = new_object_type(glob.OBJECT_TYPES[VirtualCollisionObject.__name__].meta, meta)
        source = tree_to_str(ast)

        assert meta.object_model
        await update_object_model(meta, meta.object_model)

        type_def = await hlp.run_in_executor(
            hlp.save_and_import_type_def,
            source,
            req.args.name,
            VirtualCollisionObject,
            settings.OBJECT_TYPE_PATH,
            settings.OBJECT_TYPE_MODULE,
        )
        assert issubclass(type_def, VirtualCollisionObject)
        actions = object_actions(type_def, ast)

        meta.modified = await storage.update_object_type(meta.to_object_type(source))

        glob.OBJECT_TYPES[meta.type] = ObjectTypeData(meta, type_def, actions, ast)
        add_ancestor_actions(meta.type, glob.OBJECT_TYPES)

        evt = sevts.o.ChangedObjectTypes([meta])
        evt.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))

        obj = common.SceneObject(req.args.name, req.args.name, req.args.pose)
        scene.upsert_object(obj)  # add_object_to_scene(scene, obj) would do some unnecessary checks

        evt2 = sevts.s.SceneObjectChanged(obj)
        evt2.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt2))

        # for a case when user created VCO of the type which already existed before (and was deleted)
        asyncio.create_task(unschedule_auto_remove(obj.type))

        return None



import asyncio

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.cached import UpdateableCachedScene
from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import run_in_executor
from arcor2.image import image_to_str
from arcor2.object_types.abstract import Camera
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver.helpers import ensure_write_locked
from arcor2_arserver.scene import ensure_scene_started, get_instance, update_scene_object_pose
from arcor2_arserver_data.events.common import ProcessState
from arcor2_arserver_data.rpc.camera import CalibrateCamera, CameraColorImage, CameraColorParameters
from arcor2_calibration_data import client as calib_client


async def camera_color_image_cb(req: CameraColorImage.Request, ui: WsClient) -> CameraColorImage.Response:

    glob.LOCK.scene_or_exception()

    await ensure_write_locked(req.args.id, glob.USERS.user_name(ui))

    ensure_scene_started()
    camera = get_instance(req.args.id, Camera)
    resp = CameraColorImage.Response()
    resp.data = image_to_str(camera.color_image())
    return resp


async def camera_color_parameters_cb(
    req: CameraColorParameters.Request, ui: WsClient
) -> CameraColorParameters.Response:

    glob.LOCK.scene_or_exception()

    await ensure_write_locked(req.args.id, glob.USERS.user_name(ui))

    ensure_scene_started()
    camera = get_instance(req.args.id, Camera)
    resp = CameraColorParameters.Response()
    resp.data = camera.color_camera_params
    return resp


CAMERA_CALIB = "CameraCalibration"


async def calibrate_camera(scene: UpdateableCachedScene, camera: Camera) -> None:

    assert camera.color_camera_params

    await notif.broadcast_event(ProcessState(ProcessState.Data(CAMERA_CALIB, ProcessState.Data.StateEnum.Started)))
    try:
        img = await run_in_executor(camera.color_image)
        estimated_pose = await run_in_executor(calib_client.estimate_camera_pose, camera.color_camera_params, img)
    except Arcor2Exception as e:
        await notif.broadcast_event(
            ProcessState(ProcessState.Data(CAMERA_CALIB, ProcessState.Data.StateEnum.Failed, str(e)))
        )
        logger.exception("Failed to calibrate the camera.")
        return

    await update_scene_object_pose(scene, scene.object(camera.id), estimated_pose.pose, camera)
    await notif.broadcast_event(ProcessState(ProcessState.Data(CAMERA_CALIB, ProcessState.Data.StateEnum.Finished)))


async def calibrate_camera_cb(req: CalibrateCamera.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()

    ensure_scene_started()
    camera = get_instance(req.args.id, Camera)

    # TODO check camera features / meta if it supports getting color image
    if not camera.color_camera_params:
        raise Arcor2Exception("Camera parameters not available.")

    await ensure_write_locked(req.args.id, glob.USERS.user_name(ui))

    asyncio.ensure_future(calibrate_camera(scene, camera))



import asyncio
from dataclasses import dataclass
from datetime import datetime
from typing import AsyncIterator, TypeVar

from arcor2 import helpers as hlp
from arcor2.cached import CachedScene, UpdateableCachedScene
from arcor2.clients import aio_scene_service as scene_srv
from arcor2.data.common import Parameter, Pose, SceneObject
from arcor2.data.events import Event
from arcor2.data.object_type import Models
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import CollisionObject, Generic, GenericWithPose, Robot, VirtualCollisionObject
from arcor2.object_types.utils import settings_from_params
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver.checks import check_object, scene_problems
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.common import invalidate_joints_using_object_as_parent
from arcor2_arserver.helpers import ctx_write_lock
from arcor2_arserver.lock.exceptions import CannotLock
from arcor2_arserver.object_types.utils import remove_object_type
from arcor2_arserver.objects_actions import get_object_types
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data.events.common import ShowMainScreen
from arcor2_arserver_data.events.scene import OpenScene, SceneClosed, SceneObjectChanged, SceneState
from arcor2_arserver_data.objects import ObjectTypeMeta

# TODO maybe this could be property of ARServerScene(CachedScene)?
_scene_state: SceneState = SceneState(SceneState.Data(SceneState.Data.StateEnum.Stopped))


@dataclass
class SceneProblems:
    scene_modified: datetime
    problems: list[str]
    ot_modified: dict[str, datetime]


_scene_problems: dict[str, SceneProblems] = {}
_objects_to_auto_remove: set[str] = set()


async def schedule_auto_remove(obj_type: str) -> None:
    logger.debug(f"OT {obj_type} scheduled to be autoremoved.")
    _objects_to_auto_remove.add(obj_type)


async def clear_auto_remove_schedule() -> None:
    logger.debug(f"Auto-remove schedule will be cleared. It contained: {_objects_to_auto_remove}")
    _objects_to_auto_remove.clear()


async def scheduled_to_be_auto_removed() -> set[str]:
    return _objects_to_auto_remove


async def unschedule_auto_remove(obj_type: str) -> None:

    try:
        _objects_to_auto_remove.remove(obj_type)
    except KeyError:
        pass
    else:
        logger.debug(f"OT {obj_type} unscheduled to be auto-removed.")


async def remove_scheduled() -> None:

    logger.debug(f"Going to auto-remove following types: {_objects_to_auto_remove}")
    for ot_id in _objects_to_auto_remove:
        if ot_id in glob.OBJECT_TYPES:
            asyncio.create_task(delete_if_not_used(glob.OBJECT_TYPES[ot_id].meta))
    _objects_to_auto_remove.clear()


async def delete_if_not_used(meta: ObjectTypeMeta) -> None:

    if meta.base != VirtualCollisionObject.__name__:
        logger.debug(f"{meta.type} is not a VCO!")
        return

    assert meta.object_model
    assert meta.type == meta.object_model.model().id, f"meta.type={meta.type}, model.id={meta.object_model.model().id}"

    async for scn in scenes():
        if any(scn.objects_of_type(meta.type)):
            logger.debug(f"Not auto-removing VCO {meta.type} as it is used in scene {scn.name}.")
            return

    logger.debug(f"Auto-removing VCO {meta.type} as it is not used in any scene.")

    try:
        await asyncio.gather(
            storage.delete_object_type(meta.type),
            storage.delete_model(meta.type),
            remove_object_type(meta.type),
        )
    except Arcor2Exception as e:
        logger.warn(str(e))

    del glob.OBJECT_TYPES[meta.type]

    logger.debug(f"Auto-removing {meta.type} done... {meta}")

    evtr = sevts.o.ChangedObjectTypes([meta])
    evtr.change_type = Event.Type.REMOVE
    await notif.broadcast_event(evtr)


async def save_scene(scene: None | UpdateableCachedScene = None) -> None:

    if not scene:
        scene = glob.LOCK.scene_or_exception()

    scene.modified = await storage.update_scene(scene)
    asyncio.ensure_future(notif.broadcast_event(sevts.s.SceneSaved()))
    asyncio.create_task(remove_scheduled())

    for obj_id in glob.OBJECTS_WITH_UPDATED_POSE:
        asyncio.ensure_future(invalidate_joints_using_object_as_parent(scene.object(obj_id)))
    glob.OBJECTS_WITH_UPDATED_POSE.clear()


def get_ot_modified(ots: set[str]) -> dict[str, datetime]:
    return {k: v.meta.modified for k, v in glob.OBJECT_TYPES.items() if k in ots and v.meta.modified is not None}


async def get_scene_problems(scene: CachedScene) -> None | list[str]:
    """Handle caching of scene problems."""

    assert scene.modified

    ots = scene.object_types
    await get_object_types()
    ot_modified = get_ot_modified(ots)

    if (
        scene.id not in _scene_problems
        or _scene_problems[scene.id].scene_modified < scene.modified
        or _scene_problems[scene.id].ot_modified != ot_modified
    ):

        logger.debug(f"Updating scene_problems for {scene.name}.")

        _scene_problems[scene.id] = SceneProblems(
            scene.modified,
            scene_problems(glob.OBJECT_TYPES, scene),
            get_ot_modified(ots),
        )

    # prune removed scenes
    for csi in set(_scene_problems.keys()) - await storage.get_scene_ids():
        logger.debug(f"Pruning cached problems for removed scene {csi}.")
        _scene_problems.pop(csi, None)

    sp = _scene_problems[scene.id].problems

    return sp if sp else None


async def update_scene_object_pose(
    scene: UpdateableCachedScene,
    obj: SceneObject,
    pose: None | Pose = None,
    obj_inst: None | GenericWithPose = None,
    lock_owner: None | str = None,
) -> None:
    """Performs all necessary actions when pose of an object is updated.

    :param obj:
    :param pose:
    :param obj_inst:
    :param lock_owner: if present, object is unlocked at the end of function
    :return:
    """

    if pose:
        # SceneObject pose was not updated before
        obj.pose = pose
    else:
        # SceneObject pose was already updated
        pose = obj.pose

    scene.update_modified()

    evt = SceneObjectChanged(obj)
    evt.change_type = Event.Type.UPDATE
    await notif.broadcast_event(evt)

    glob.OBJECTS_WITH_UPDATED_POSE.add(obj.id)

    if scene_started():

        if obj_inst is None:
            obj_inst = get_instance(obj.id, GenericWithPose)

        assert pose is not None

        # Object pose is property that might call scene service - that's why it has to be called using executor.
        await hlp.run_in_executor(setattr, obj_inst, "pose", pose)

    if lock_owner:
        await glob.LOCK.write_unlock(obj.id, lock_owner, True)


async def set_scene_state(state: SceneState.Data.StateEnum, message: None | str = None) -> None:

    global _scene_state
    _scene_state = SceneState(SceneState.Data(state, message))
    asyncio.create_task(notif.broadcast_event(_scene_state))


def get_scene_state() -> SceneState:
    return _scene_state


def scene_started() -> bool:
    return _scene_state.data.state == SceneState.Data.StateEnum.Started


def can_modify_scene() -> None:
    """Raises exception if modifications to scene/project are not possible."""

    if _scene_state.data.state != SceneState.Data.StateEnum.Stopped:
        raise Arcor2Exception("Modifications can be only done offline.")


def ensure_scene_started() -> None:
    """Raises exception if scene is not started."""

    if _scene_state.data.state != SceneState.Data.StateEnum.Started:
        raise Arcor2Exception("Scene offline.")


async def notify_scene_opened(evt: OpenScene) -> None:

    await notif.broadcast_event(evt)
    ss = get_scene_state()
    assert ss.data.state == ss.Data.StateEnum.Stopped
    await notif.broadcast_event(ss)


async def scenes() -> AsyncIterator[CachedScene]:

    for scene_id in await storage.get_scene_ids():
        yield await storage.get_scene(scene_id)


async def scene_names() -> set[str]:
    return {scene.name for scene in (await storage.get_scenes())}


async def notify_scene_closed(scene_id: str) -> None:

    assert get_scene_state().data.state == SceneState.Data.StateEnum.Stopped

    await notif.broadcast_event(SceneClosed())
    glob.MAIN_SCREEN = ShowMainScreen.Data(ShowMainScreen.Data.WhatEnum.ScenesList)
    await notif.broadcast_event(
        ShowMainScreen(data=ShowMainScreen.Data(ShowMainScreen.Data.WhatEnum.ScenesList, scene_id))
    )


async def add_object_to_scene(scene: UpdateableCachedScene, obj: SceneObject, dry_run: bool = False) -> None:
    """

    :param obj:
    :param add_to_scene: Set to false to only create object instance and add its collision model (if any).
    :return:
    """

    check_object(glob.OBJECT_TYPES, scene, obj, new_one=True)

    if dry_run:
        return None

    scene.upsert_object(obj)
    logger.debug(f"Object {obj.id} ({obj.type}) added to scene.")


async def create_object_instance(obj: SceneObject, overrides: None | list[Parameter] = None) -> None:

    obj_type = glob.OBJECT_TYPES[obj.type]

    # settings -> dataclass
    assert obj_type.type_def
    logger.debug(
        f"Creating instance of {obj_type.type_def.__name__} with name {obj.name}. "
        f"Parameters: {obj.parameters}, overrides: {overrides}."
    )
    settings = settings_from_params(obj_type.type_def, obj.parameters, overrides)

    assert obj_type.type_def is not None

    try:

        try:

            # the order must be from specific to the generic types
            if issubclass(obj_type.type_def, Robot):
                assert obj.pose is not None
                glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(
                    obj_type.type_def, obj.id, obj.name, obj.pose, settings
                )
            elif issubclass(obj_type.type_def, CollisionObject):
                assert obj.pose is not None
                coll_model: None | Models = None
                if obj_type.meta.object_model:
                    coll_model = obj_type.meta.object_model.model()

                glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(
                    obj_type.type_def, obj.id, obj.name, obj.pose, coll_model, settings
                )
            elif issubclass(obj_type.type_def, GenericWithPose):
                assert obj.pose is not None

                glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(
                    obj_type.type_def, obj.id, obj.name, obj.pose, settings
                )
            elif issubclass(obj_type.type_def, Generic):
                assert obj.pose is None
                glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(
                    obj_type.type_def, obj.id, obj.name, settings
                )

            else:
                raise Arcor2Exception("Object type with unknown base.")

        except (TypeError, ValueError) as e:  # catch some most often exceptions
            raise Arcor2Exception("Unhandled error.") from e

    except Arcor2Exception as e:
        # make the exception a bit more user-friendly by including the object's name
        raise Arcor2Exception(f"Failed to initialize {obj.name}. {str(e)}") from e

    return None


async def open_scene(scene_id: str) -> None:

    scene = await storage.get_scene(scene_id)

    if sp := await get_scene_problems(scene):  # this also refreshes ObjectTypes / calls get_object_types()
        logger.warning(f"Scene {scene.name} can't be opened due to the following problem(s)...")
        for spp in sp:
            logger.warning(spp)
        raise Arcor2Exception("Scene has some problems.")

    glob.LOCK.scene = UpdateableCachedScene(scene)


def get_robot_instance(obj_id: str) -> Robot:  # TODO remove once https://github.com/python/mypy/issues/5374 is solved

    robot_inst = get_instance(obj_id, Robot)  # type: ignore
    assert isinstance(robot_inst, Robot)
    return robot_inst


T = TypeVar("T", bound=Generic)


# TODO "thanks" to https://github.com/python/mypy/issues/3737, `expected_type: type[T] = Generic` can't be used
def get_instance(obj_id: str, expected_type: type[T]) -> T:

    assert scene_started()

    try:
        inst = glob.SCENE_OBJECT_INSTANCES[obj_id]
    except KeyError:
        raise Arcor2Exception("Unknown object ID.")

    if not isinstance(inst, expected_type):
        raise Arcor2Exception(f"{inst.name} is not of {expected_type.__name__} type.")

    return inst


async def cleanup_object(obj: Generic) -> None:

    try:
        await hlp.run_in_executor(obj.cleanup)
    except Arcor2Exception as e:
        # make the exception a bit more user-friendly by including the object's name
        raise Arcor2Exception(f"Failed to cleanup {obj.name}. {str(e)}") from e


async def stop_scene(scene: CachedScene, message: None | str = None, already_locked: bool = False) -> None:
    """Destroys scene object instances."""

    async def _stop_scene() -> None:

        await set_scene_state(SceneState.Data.StateEnum.Stopping, message)

        try:
            await scene_srv.stop()
        except Arcor2Exception as e:
            logger.exception("Failed to go offline.")
            await set_scene_state(SceneState.Data.StateEnum.Started, str(e))
            return

        try:
            await asyncio.gather(*[cleanup_object(obj) for obj in glob.SCENE_OBJECT_INSTANCES.values()])
        except Arcor2Exception as e:
            logger.exception("Exception occurred while cleaning up objects.")
            await set_scene_state(SceneState.Data.StateEnum.Stopped, str(e))
        else:
            await set_scene_state(SceneState.Data.StateEnum.Stopped)

        glob.SCENE_OBJECT_INSTANCES.clear()
        glob.PREV_RESULTS.clear()

    if already_locked:

        logger.info(f"Stopping the {scene.name} scene after unsuccessful start.")

        assert await glob.LOCK.is_write_locked(glob.LOCK.SpecialValues.SCENE, glob.LOCK.Owners.SERVER)
        assert (glob.LOCK.project is not None) == await glob.LOCK.is_write_locked(
            glob.LOCK.SpecialValues.PROJECT, glob.LOCK.Owners.SERVER
        )

        await _stop_scene()
    else:

        to_lock = [glob.LOCK.SpecialValues.SCENE]

        if glob.LOCK.project:
            to_lock.append(glob.LOCK.SpecialValues.PROJECT)

        try:
            async with ctx_write_lock(to_lock, glob.LOCK.Owners.SERVER):
                logger.info(f"Stopping the {scene.name} scene.")
                await _stop_scene()
        except CannotLock:
            logger.warning(f"Failed attempt to stop the scene. Can't lock {to_lock}.")
            return
        except Arcor2Exception as e:
            logger.error(f"Failed to stop the scene. {str(e)}")
            return

    assert not scene_started()
    assert not await scene_srv.started()

    logger.info("Scene stopped.")


async def start_scene(scene: CachedScene) -> None:
    """Creates instances of scene objects."""

    async def _start_scene() -> bool:

        logger.info(f"Starting the {scene.name} scene.")

        await set_scene_state(SceneState.Data.StateEnum.Starting)

        # in order to prepare a clear environment
        try:
            # stop deletes all configurations and clears all collisions
            await scene_srv.stop()
        except Arcor2Exception:
            logger.exception("Failed to prepare for start.")
            await set_scene_state(SceneState.Data.StateEnum.Stopped, "Failed to prepare for start.")
            return False

        object_overrides: dict[str, list[Parameter]] = {}

        if glob.LOCK.project:
            object_overrides = glob.LOCK.project.overrides

        # object initialization could take some time - let's do it in parallel
        tasks = [
            asyncio.ensure_future(
                create_object_instance(obj, object_overrides[obj.id] if obj.id in object_overrides else None)
            )
            for obj in scene.objects
        ]

        try:
            await asyncio.gather(*tasks)
        except Arcor2Exception as e:
            for t in tasks:
                t.cancel()  # TODO maybe it would be better to let them finish?
            logger.exception("Failed to create instances.")
            await stop_scene(scene, str(e), already_locked=True)
            return False

        try:
            await scene_srv.start()
        except Arcor2Exception as e:
            logger.exception("Failed to go online.")
            await stop_scene(scene, str(e), already_locked=True)
            return False

        await set_scene_state(SceneState.Data.StateEnum.Started)
        return True

    to_lock = [glob.LOCK.SpecialValues.SCENE]

    if glob.LOCK.project:
        to_lock.append(glob.LOCK.SpecialValues.PROJECT)

    try:
        async with ctx_write_lock(to_lock, glob.LOCK.Owners.SERVER):
            ret = await _start_scene()
    except CannotLock:
        logger.warning(f"Failed attempt to start the scene. Can't lock {to_lock}.")
        return
    except Arcor2Exception as e:
        logger.error(f"Failed to start the scene. {str(e)}")

    assert ret == scene_started()
    assert ret == await scene_srv.started()

    if ret:
        logger.info("Scene started. Enjoy!")



from __future__ import annotations

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2_arserver.lock.common import ObjIds, obj_ids_to_list
from arcor2_arserver.lock.exceptions import CannotUnlock, LockingException


class LockEventData:
    __slots__ = "obj_ids", "owner", "lock", "client"

    def __init__(self, obj_ids: ObjIds, owner: str, lock: bool = False, client: None | WsClient = None):
        self.obj_ids = obj_ids_to_list(obj_ids)
        self.owner = owner
        self.lock = lock
        self.client = client


class LockedObject:
    __slots__ = "read", "write", "tree"

    def __init__(self) -> None:

        # object id: owners
        self.read: dict[str, list[str]] = {}
        self.write: dict[str, str] = {}

        self.tree: bool = False

    def __repr__(self) -> str:
        return f"LockedObject(read={self.read}, write={self.write}, tree={self.tree})"

    def read_lock(self, obj_id: str, owner: str) -> bool:
        """Perform all necessary check if object can be locked and locks it."""

        if self.tree or obj_id in self.write:
            return False

        if obj_id in self.read:  # already locked
            self.read[obj_id].append(owner)
        else:
            self.read[obj_id] = [owner]
        return True

    def read_unlock(self, obj_id: str, owner: str) -> None:
        """Perform all necessary check if object can be unlocked and unlocks
        it."""

        if obj_id not in self.read:
            raise CannotUnlock(f"{obj_id} is not read locked by {owner}.")

        if owner not in self.read[obj_id]:
            raise CannotUnlock(f"{obj_id} lock is not owned by {owner}.")

        if len(self.read[obj_id]) > 1:
            self.read[obj_id].remove(owner)
        else:
            del self.read[obj_id]

    def write_lock(self, obj_id: str, owner: str, lock_tree: bool) -> bool:
        """Perform all necessary check if object can be locked and locks it."""

        if self.tree or obj_id in self.write or obj_id in self.read:
            return False

        if lock_tree and (self.read or self.write):
            return False

        self.write[obj_id] = owner
        self.tree = lock_tree
        return True

    def write_unlock(self, obj_id: str, owner: str) -> None:
        """Perform all necessary check if object can be unlocked and unlocks
        it."""

        if obj_id not in self.write:
            raise CannotUnlock(f"{obj_id} is not write locked by {owner}.")

        if owner != self.write[obj_id]:
            raise CannotUnlock(f"{obj_id} lock is not owned by {owner}.")

        if self.tree:
            self.tree = False
        del self.write[obj_id]

    def is_empty(self) -> bool:
        """Check if there are any locked objects in current tree."""

        return not self.read and not self.write

    def check_upgrade(self, obj_id: str, owner: str) -> None:
        """Check if lock can be upgraded to whole locked tree."""

        raise_msg = f"{obj_id} lock is not owned by {owner}."
        if obj_id not in self.write:
            raise LockingException(raise_msg)

        if len(self.write) > 1:
            raise LockingException(raise_msg)

        if owner != self.write[obj_id]:
            raise LockingException(raise_msg)

        if self.tree:
            raise LockingException(f"Nothing to upgrade for {obj_id} and owner {owner}.")

    def check_downgrade(self, obj_id: str, owner: str) -> None:
        """Check if lock can be downgraded to single object lock."""

        raise_msg = f"{obj_id} lock is not owned by {owner}."
        if obj_id not in self.write:
            raise LockingException(raise_msg)

        if owner != self.write[obj_id]:
            raise LockingException(raise_msg)

        if not self.tree:
            raise LockingException(f"Nothing to downgrade for {obj_id} and owner {owner}.")



import asyncio

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.exceptions import Arcor2Exception
from arcor2_arserver import globals as glob
from arcor2_arserver.execution import build_and_upload_package, run_temp_package
from arcor2_arserver.helpers import ctx_write_lock
from arcor2_arserver_data import rpc


async def build_project_cb(req: rpc.b.BuildProject.Request, ui: WsClient) -> rpc.b.BuildProject.Response:
    """Builds project and uploads resulting package to the execution unit.

    :param req:
    :param ui:
    :return:
    """

    async with ctx_write_lock(req.args.project_id, glob.USERS.user_name(ui)):
        package_id = await build_and_upload_package(req.args.project_id, req.args.package_name)

        resp = rpc.b.BuildProject.Response()
        resp.data = resp.Data(package_id)
        return resp


async def temporary_package_cb(req: rpc.b.TemporaryPackage.Request, ui: WsClient) -> None:

    async with glob.LOCK.get_lock():

        project = glob.LOCK.project_or_exception()

        if project.has_changes:
            raise Arcor2Exception("Project has unsaved changes.")

        package_id = await build_and_upload_package(project.id, f"Temporary package for project '{project.name}'.")

        if req.args:
            paused = req.args.start_paused
            breakpoints = req.args.breakpoints
        else:
            paused = False
            breakpoints = None

        asyncio.ensure_future(run_temp_package(package_id, paused, breakpoints))
        return None
